#!/bin/csh
# usage : DO_tbl* directory {z1 z2}
#
set do_tbl=`basename $0`
#
if ($#argv != 1 && $#argv != 3) then
  stderr usage : $do_tbl directory \{z1 z2\} ; exit 1
endif
#
set dir=$argv[1]
if (!(-d $dir)) then
  stderr $dir : directory not found. ; exit 1
endif
#
set log=${dir}/output.log
if (!(-e $log)) then
  stderr $log : log file not found. ; exit 1
endif
#
set dark=${dir}/d.img
if (!(-e $dark)) then
  set dark=${dir}/dark.img
  if (!(-e $dark)) then
    stderr $dark : dark-image file not found. ; exit 1
  endif
endif
#
alias GREP_SED 'grep \^q\[0-9\]\*\[0-9\].img | sed -e s/\^q//g -e s/.img\$//g'
set awk=('BEGIN { I=E=0 }' \
         'NF==1 { Z[$1+0]=$1 }' \
         'NF==4 { if ((z=Z[$1+0])!="")' \
                   'if ($4==0) { Z0[I]=z; T0[I]=$2; ++I }' \
                   'else if ($4==1)' \
                     'if (++E==1) { Z1=Z2=z; T1=T2=$2; A1=A2=$3 }' \
                     'else if ($3<A1) { Z1=z; T1=$2; A1=$3 }' \
                     'else if ($3>A2) { Z2=z; T2=$2; A2=$3 }}' \
         'function abs(f) { return (f<0)?-f:f }' \
         'function end(t,z,a) {' \
           'D=abs(t-T0[i0=0]);' \
           'for (i=1; i<I; i++) if ((d=abs(t-T0[i]))<D) { D=d; i0=i }' \
           'printf "'${dir}'/q%s.img\n'${dir}'/q%s.img\n%s\n",Z0[i0],z,a }' \
         'END { if (I>=1 && E>=2) { end(T1,Z1,A1); end(T2,Z2,A2) }}')
set it=(`( ls $dir | GREP_SED ; cat $log | tr -d \\r ) | awk "$awk"`)
if ($#it != 6) then
  stderr ${dir}/q\[0-9\]\*\[0-9\].img : I0- and/or I-image file\(s\) not found.
  exit 1
endif
#
stderr D000 = $dark
stderr I000 = $it[1]
stderr T000 = $it[2] at angle = $it[3]
stderr D180 = $dark
stderr I180 = $it[4]
stderr T180 = $it[5] at angle = $it[6]
#
hp2DO $dark $it[1-2] $dark $it[4-5] `echo $argv | awk 'NF==3 { print $2,$3 }'`
