#!/bin/csh
# usage : [t,p]xy_tif* directory angle {{base step} BPS [t,p]xy.tif}
#
set xy_tif=`basename $0`
if ($xy_tif !~ [t,p]xy_tif*) then
  stderr $xy_tif : unknown command. ; exit 1
endif
set xy=`echo $xy_tif | cut -d_ -f1`
#
if ($#argv != 2 && $#argv != 4 && $#argv != 6) then
  stderr usage : $xy_tif directory angle \{\{base step\} BPS ${xy}.tif\}
  exit 1
endif
#
set dir=$argv[1]
if (!(-d $dir)) then
  stderr $dir : directory not found. ; exit 1
endif
#
set log=${dir}/output.log
if (!(-e $log)) then
  stderr $log : log file not found. ; exit 1
endif
#
set dark=${dir}/d.img
if (!(-e $dark)) then
  set dark=${dir}/dark.img
  if (!(-e $dark)) then
    stderr $dark : dark-image file not found. ; exit 1
  endif
endif
#
alias GREP_SED 'grep \^q\[0-9\]\*\[0-9\].img | sed -e s/\^q//g -e s/.img\$//g'
set awk=("BEGIN { I=E=0; A=$argv[2] }" \
         'NF==1 { Z[$1+0]=$1 }' \
         'function abs(f) { return (f<0)?-f:f }' \
         'NF==4 { if ((z=Z[$1+0])!="")' \
                   'if ($4==0) { Z0[I]=z; T0[I]=$2; ++I }' \
                   'else if ($4==1)' \
                     'if (++E==1) { Ze=z; Te=$2; D=abs(A-(Ae=$3)) }' \
                     'else if ((d=abs(A-$3))<D) { Ze=z; Te=$2; D=d; Ae=$3 }}' \
         'END { if (I>=1 && E>=1) {' \
                 'D=abs(Te-T0[i0=0]);' \
                 'for (i=1; i<I; i++) if ((d=abs(Te-T0[i]))<D) { D=d; i0=i }' \
                 'printf "'${dir}'/q%s.img\n'${dir}'/q%s.img\n%s\n",' \
                        'Z0[i0],Ze,Ae }}')
set it=(`( ls $dir | GREP_SED ; cat $log | tr -d \\r ) | awk "$awk"`)
if ($#it != 3) then
  stderr ${dir}/q\[0-9\]\*\[0-9\].img : I0- and/or I-image file\(s\) not found.
  exit 1
endif
#
stderr d.img = $dark
stderr i.img = $it[1]
stderr t.img = $it[2] at angle = $it[3]
#
if ($#argv != 2) hp3$xy $dark $it[1-2] $argv[3-]
