
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cell.h"
#include "rif.h"
#include "sif.h"

extern void	Error();

#define GS_LEN	2048
#define FPRINTF	(void)fprintf

static void	AddLabel(TIFF,Nx,Ny,data,GS,name,size,pv,h,v,label)
char		*TIFF,*GS,*name,*label;
int		Nx,Ny,pv,h,v;
Cell		**data;
double		size;
{
	char	gs[GS_LEN],*ptr;
	FILE	*file;
	Cell	**mask;
	int	y,x;

	(void)sprintf(gs,
		      "%s -q -r72 -g%dx%d -sDEVICE=tifflzw -sOutputFile=%s -",
		      GS,Nx,Ny,TIFF);
	if ((file=popen(gs,"w"))==NULL) Error("gs initialization error.");

	FPRINTF(file,"/%s findfont %lf scalefont setfont\n"
		     "0 [\n(",name,size);
	for (ptr=label; *ptr!='\0'; ptr++)
	    switch(*ptr) {
		case '\\' :
		    if (ptr[1]=='n') {
			(void)fputs(")\n(",file); ++ptr; break;
		    }
		case '(' :
		case ')' : (void)fputc('\\',file);
		default : (void)fputc((int)*ptr,file);
	    }

	(void)fputs(")\n] {\n",file);
	switch(h) {
	    case 'L' :
		FPRINTF(file,"%lf 0.5 mul\n",size); break;
	    case 'C' :
		FPRINTF(file,"dup stringwidth pop %d sub 2 div neg\n",
			     Nx); break;
	    case 'R' :
		FPRINTF(file,"dup stringwidth pop %d sub %lf 0.5 mul add neg\n",
			     Nx,size); break;
	}
	FPRINTF(file,"3 -1 roll dup %lf mul\n",size);
	switch(v) {
	    case 'U' :
		FPRINTF(file,"%d %lf 1.375 mul sub exch sub\n",Ny,size); break;
	    case 'D' :
		FPRINTF(file,"%lf 0.625 mul add\n",size); break;
	}
	(void)fputs("exch 1 add 4 1 roll\n"
		    "newpath moveto show\n"
		    "} forall pop showpage\n",file);

	if (pclose(file)==(-1)) Error("gs termination error.");

	ReadImageFile(TIFF,NULL,NULL,NULL,&mask,NULL);

	for (y=0; y<Ny; y++) {
	    for (x=0; x<Nx; x++) if (mask[y][x]) data[y][x]=(Cell)pv;

	    free(mask[y]);
	}
	free(mask);
}

static void	StoreTIFF(TIFF,Nx,Ny,data)
char		*TIFF;
int		Nx,Ny;
Cell		**data;
{
	int	y,x;
	Cell	pv=0;

	for (y=0; y<Ny; y++)
	for (x=0; x<Nx; x++) if (data[y][x]>pv) pv=data[y][x];

	StoreImageFile(TIFF,Nx,Ny,(pv<2)?1:(pv<4)?2:(pv<16)?4:(pv<256)?8:16,
		       data,NULL);

	for (y=0; y<Ny; y++) free(data[y]); free(data);
}

static double	GetSize(Size)
char		*Size;
{
	double	size=atof(Size);

	return (size>=0.0)?size:0.0;
}

static int	GetPV(PV)
char		*PV;
{
	int	pv=atoi(PV);

	return (pv>=0 && pv<=(int)((Cell)~0))?pv:-1;
}

static int	GetHV(HV,h,v)
char		*HV;
int		*h,*v;
{
	return ((*h=(int)HV[0]&223)=='L' || *h=='C' || *h=='R') &&
	       ((*v=(int)HV[1]&223)=='U' || *v=='D');
}

#define STR_LEN	4096
#define DELIM	" \t\n"

int	main(argc,argv)
int	argc;
char	**argv;
{
	char	*GS,desc[STR_LEN],*name,str[STR_LEN],*tok,label[STR_LEN],*TIFF,
		*gs="gs",
		*bd="bad description (skipping) :\n%s",
		*bfs="bad font size.",
		*bfpv="bad font pixel value.",
		*blp="bad label position ([L,C,R][U,D]).";
	int	Nx,Ny,pv,h,v,a;
	double	size;
	Cell	**data;

	if (argc!=2 && argc!=5 && argc<6)
	    Error("usage : add_label TIFF\n"
		  "        add_label      fontName size PV [L,C,R][U,D]\n"
		  "        add_label TIFF fontName size PV [L,C,R][U,D] label");

	if ((GS=getenv("GS"))==NULL) GS=gs;

switch(argc) {
    case 2 :
	ReadImageFile(argv[1],&Nx,&Ny,NULL,&data,NULL);

	while (fgets(desc,STR_LEN,stdin)!=NULL)
	    if ((name=strtok(strcpy(str,desc),DELIM))!=NULL)
	    if ((tok=strtok(NULL,DELIM))==NULL || (size=GetSize(tok))<=0.0 ||
		(tok=strtok(NULL,DELIM))==NULL || (pv=GetPV(tok))<0 ||
		(tok=strtok(NULL,DELIM))==NULL || !GetHV(tok,&h,&v) ||
		(tok=strtok(NULL,DELIM))==NULL)
		FPRINTF(stderr,bd,desc);
	    else {
		(void)strcpy(label,tok);
		while ((tok=strtok(NULL,DELIM))!=NULL)
		    (void)strcat(strcat(label," "),tok);

		AddLabel(argv[1],Nx,Ny,data,GS,name,size,pv,h,v,label);
	    }

	StoreTIFF(argv[1],Nx,Ny,data);

	break;
    case 5 :
	if ((size=GetSize(argv[2]))<=0.0) Error(bfs);

	if ((pv=GetPV(argv[3]))<0) Error(bfpv);

	if (!GetHV(argv[4],&h,&v)) Error(blp);

	while (fgets(desc,STR_LEN,stdin)!=NULL)
	    if ((TIFF=strtok(strcpy(str,desc),DELIM))!=NULL)
	    if ((tok=strtok(NULL,DELIM))==NULL)
		FPRINTF(stderr,bd,desc);
	    else {
		(void)strcpy(label,tok);
		while ((tok=strtok(NULL,DELIM))!=NULL)
		    (void)strcat(strcat(label," "),tok);

		ReadImageFile(TIFF,&Nx,&Ny,NULL,&data,NULL);
		AddLabel(TIFF,Nx,Ny,data,GS,argv[1],size,pv,h,v,label);
		StoreTIFF(TIFF,Nx,Ny,data);
	    }

	break;
    default :
	if ((size=GetSize(argv[3]))<=0.0) Error(bfs);

	if ((pv=GetPV(argv[4]))<0) Error(bfpv);

	if (!GetHV(argv[5],&h,&v)) Error(blp);

	(void)strcpy(label,argv[6]);
	for (a=7; a<argc; a++) (void)strcat(strcat(label," "),argv[a]);

	ReadImageFile(argv[1],&Nx,&Ny,NULL,&data,NULL);
	AddLabel(argv[1],Nx,Ny,data,GS,argv[2],size,pv,h,v,label);
	StoreTIFF(argv[1],Nx,Ny,data);

	break;
}
	return 0;
}
