
#include <stdlib.h>
#include <math.h>
#include "cbp.h"

#ifndef	Int
#define Int	int
#endif

#ifndef	P_BITS
#define P_BITS	12
#endif

#ifndef	R_BITS
#define R_BITS	9
#endif

#ifndef	Q_BITS
#define Q_BITS	(31-P_BITS-R_BITS)
#endif

#ifndef	S_BITS
#define S_BITS	(30-P_BITS-R_BITS)
#endif

#if	defined(__x86_64__) || defined(__ia64__)
#define	__64BIT_DATABUS
#endif

double	Ramachandran(i)
int	i;
{
	return (i==0)?1.0/4.0:(i&1)?-1.0/(M_PI*M_PI*(double)i*(double)i):0.0;
}

double	Shepp(i)
int	i;
{
	return 2.0/(M_PI*M_PI*(1.0-4.0*(double)i*(double)i));
}

double	Chesler(i)
int	i;
{
	if (i==0)
	    return 1.0/8.0-1.0/(2.0*M_PI*M_PI);
	else if (i==1 || i==(-1))
	    return 1.0/16.0-1.0/(2.0*M_PI*M_PI);
	else if (i&1)
	    return -1.0/(2.0*M_PI*M_PI*(double)i*(double)i);
{
	double	i2=(double)i*(double)i,i21=i2-1.0;

	return -(i2+1.0)/(2.0*M_PI*M_PI*i21*i21);
}}

typedef struct {
		double	x,y;
	} Vector;

static void	FFT(L,E,Z,sign)
int		L,sign;
Vector		*E,*Z;
{
	int	j,k,m,n,p,q,L2=L>>1;
	Vector	z,*e,*a,*b;
	double	x,y,w;

	j=L-1;
	for (k=L-2; k>0; k--) {
	    for (m=L2; (j^=m)&m; m>>=1) ;

	    if (j>k) {
		z=Z[j]; Z[j]=Z[k]; Z[k]=z;
	    }
	}
	for (j=2, k=1, m=L2, n=L; n>1; n=m, m>>=1, k=j, j<<=1)
	    for (e=E, p=0; p<k; p++, e+=n) {
		x=e->x; y=e->y*(double)sign;
		for (a=Z+p, q=0; q<m; q++, a+=j) {
		    z=(*(b=a+k)); b->x=a->x-(w=z.x*x-z.y*y); a->x+=w;
				  b->y=a->y-(w=z.x*y+z.y*x); a->y+=w;
		}
	    }
}

static int	N,M,*H0,*dH;
static Float	**P,**Q,**F;
static Vector	*E,*Z,*G;

#define POW2(N)		(1<<(int)ceil(log((double)(N))/log(2.0)))
#define ALLOC(type,noe)	(type *)malloc(sizeof(type)*(size_t)(noe))

Float	**InitCBP(Ni,Mi)
int	Ni,Mi;
{
	int	m,n,L=POW2(Ni),L2=L<<1;
	double	a,da=M_PI/(double)L2;

	if ((P =ALLOC(Float *,Mi   ))==NULL ||
	    (*P=ALLOC(Float  ,Mi*Ni))==NULL ||
	    (E =ALLOC(Vector ,L2   ))==NULL ||
	    (Z =ALLOC(Vector ,L2   ))==NULL ||
	    (G =ALLOC(Vector ,L2   ))==NULL ||
	    (Q =ALLOC(Float *,Mi   ))==NULL ||
	    (*Q=ALLOC(Float  ,Mi*Ni))==NULL ||
	    (H0=ALLOC(int    ,Ni   ))==NULL ||
	    (dH=ALLOC(int    ,Ni   ))==NULL ||
	    (F =ALLOC(Float *,Ni   ))==NULL ||
	    (*F=ALLOC(Float  ,Ni*Ni))==NULL) return NULL;

	for (m=1; m<Mi; m++) {
	    P[m]=P[m-1]+Ni; Q[m]=Q[m-1]+Ni;
	}
	for (n=1; n<Ni; n++) F[n]=F[n-1]+Ni;

	for (n=0; n<L2; n++) {
	    a=da*(double)n; E[n].x=cos(a); E[n].y=sin(a);

	    Z[n].x=Filter(n-L); Z[n].y=0.0;
	}
	FFT(L2,E,Z,-1);

	for (n=0; n<L2; n++) G[n]=Z[n];

	N=Ni; M=Mi; return P;
}

Float	**CBP(dr,r0,t0)
double	dr,r0,t0;
{
	int	v1,v2,v,h,m,l,n,L=POW2(N),L2=L<<1,N1=N-1;
	double	y,x,q0,t,
		N12=(double)N1/2.0,R=N12-fabs(N12+r0),R2=R*R,
		Q0=0.0,dt=M_PI/(double)M;
	Float	*p,*q,**Fv,*Fh;
	Int	s,c,rv,**fv,rh,*fh,
		**f=(Int **)F,R0=(Int)floor(ldexp(r0-1.0,R_BITS+S_BITS)+0.5);
#ifndef	__64BIT_DATABUS
	Int	*qi=(Int *)Z+1;
	int	n1;
#else
	struct Vector {
		Int	x,y;
	} *qi=(struct Vector *)Z+1;
#endif

#ifndef	__INTEL_COMPILER
	v1=(int)ceil(N12-R); v2=(int)(N12+R);
	for (y=N12-(double)(v=v1); v<=v2; v++, y-=1.0) {
	    x=sqrt(R2-y*y); dH[v]=(int)(N12+x)-(H0[v]=(int)ceil(N12-x));
	}
	for (v=0; v<N; v++)
	for (h=0; h<N; h++) f[v][h]=0;
#else
	v1=N; v2=(-1);
	for (y=N12, v=0; v<N; v++, y-=1.0) {
	    if (fabs(y)<=R) {
		x=sqrt(R2-y*y); dH[v]=(int)(N12+x)-(H0[v]=(int)ceil(N12-x));

		if (v<v1) v1=v; v2=v;
	    }
	    for (h=0; h<N; h++) f[v][h]=0;
	}
#endif

	for (m=0; m<M; m++) {
	    for (	 l=0; l<L;	l++)   Z[l].x=	    Z[l].y=0.0;
	    for (p=P[m], n=0; n<N; n++, l++) { Z[l].x=p[n]; Z[l].y=0.0; }
	    for (	    ; n<L; n++, l++)   Z[l].x=	    Z[l].y=0.0;

	    FFT(L2,E,Z,-1);

	    for (n=0; n<L2; n++) {
		x=Z[n].x;
		y=Z[n].y; Z[n].x=x*G[n].x-y*G[n].y;
			  Z[n].y=x*G[n].y+y*G[n].x;
	    }
	    FFT(L2,E,Z, 1);

	    for (q=Q[m], n=0; n<N; n++)
		if ((q0=fabs(q[n]=Z[n].x))>Q0) Q0=q0;
	}
	q0=ldexp(1.0/Q0,Q_BITS);

#ifndef	__64BIT_DATABUS
	qi[-1]=qi[N]=0;
#else
	qi[-1].x=qi[N].x=qi[N].y=0;
#endif
	for (t=t0, m=0; m<M; m++, t+=dt) {
	    q=Q[m];
#ifndef	__64BIT_DATABUS
	    for (n=0; n<N; n++) qi[n]=(Int)floor(q[n]*q0+0.5);
#else
	    qi+=N; q+=N;
	    for (n=0; n<N; n++) {
		--qi; --q; qi->y=qi[1].x-(qi->x=(Int)floor(*q*q0+0.5));
	    }
	    qi[-1].y=qi->x;
#endif
	    s=(Int)floor(ldexp(sin(t),R_BITS+S_BITS)+0.5);
	    c=(Int)floor(ldexp(cos(t),R_BITS+S_BITS)+0.5);

	    rv=(N1*(s-c)+1)/2-v1*s-R0; fv=f+v1;
	    for (v=v1; v<=v2; v++, rv-=s, fv++) {
		h=H0[v]; rh=h*c+rv; fh=(*fv)+h;
		for (h=dH[v]; h>=0; h--, rh+=c, fh++) {
#ifndef	__64BIT_DATABUS
	n=(n1=(int)(rh>>R_BITS+S_BITS))-1;
	*fh+=(qi[n]*(1<<R_BITS)+(qi[n1]-qi[n])*((rh>>S_BITS)&((1<<R_BITS)-1)));
#else
	n=(int)(rh>>R_BITS+S_BITS)-1;
	*fh+=(qi[n].x*(1<<R_BITS)+qi[n].y*((rh>>S_BITS)&((1<<R_BITS)-1)));
#endif
		}
	    }
	}
	q0=ldexp(dt*Q0/(dr*(double)L2),-(Q_BITS+R_BITS));
	for (Fv=(Float **)(fv=f+N), v=0; v<N; v++)
	for (Fh=(*(--Fv))+N,
	     fh=(*(--fv))+N, h=0; h<N; h++) *(--Fh)=q0*(double)*(--fh);

	return F;
}

void	TermCBP()
{
	free(*F); free(F); free(dH); free(H0);
	free(*Q); free(Q); free(G); free(Z); free(E);
	free(*P); free(P);
}
