
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "cbp.h"
#include "cell.h"
#include "sif.h"

#define time	tick	/* to prevent error of lint */

static void	Error(head,tail)
char		*head,*tail;
{
	if (*head!='\0') (void)fprintf(stderr,"%s : ",head);

	(void)fputs(tail,stderr); (void)fputc('\n',stderr); exit(1);
}

#define ALLOC(type,noe)	(type *)malloc(sizeof(type)*(size_t)(noe))

#define SWAP(word,noe)	\
{ \
	short	tmp0=127; \
	char	*tmp1,tmp2; \
\
	if (*((char *)&tmp0)!=127) \
	    for (tmp1=(char *)(word+noe-1); tmp1>=(char *)word; tmp1-=2) { \
		tmp2=tmp1[1]; tmp1[1]=tmp1[0]; tmp1[0]=tmp2; \
	    } \
}

typedef unsigned short	Word;
typedef struct {
		union {
			struct {
				char	*path;
				long	offset;
			} f;
			Word	*memory;
		} u;
		double	time;
	} Table;
typedef struct {
#ifdef	OM
		union {
			struct {
				char	*path;
				long	offset;
			} f;
			Word	*memory;
		} u;
#else
		char	*path;
#ifdef	OF
		union {
			long	offset;
			FILE	*file;
		} u;
#else
		long	offset;
#endif
#endif
		double	time,Rd1,Rd2,Ri1,Ri2;
		Word	*Md1,*Md2,*Mi1,*Mi2;
	} TableT;

static int	Nh,Nv,first=1;
static double	time1,time2;

#define STR_LEN	1024

static void	Check(tPath,Nt,di,T)
char		*tPath;
int		*Nt,di;
Table		**T;
{
	FILE	*tFile,*file;
	char	line[STR_LEN],path[STR_LEN],
		*tmae="table memory allocation error.";
	double	time;

	struct List {
			char		*path;
			long		offset;
			double		time;
			struct List	*next;
		} *l,*next,**L=(&next);

	short	IH[7];

	if (*tPath=='-')
	    tFile=stdin;
	else
	    if ((tFile=fopen(tPath,"r"))==NULL) Error(tPath,"file not found.");

	for (*Nt=0; fgets(line,STR_LEN,tFile)!=NULL &&
		    sscanf(line,"%s %lf",path,&time)==2; (*Nt)++) {
	    if ((*L=l=ALLOC(struct List,1))==NULL ||
		(l->path=strdup(path))==NULL)
		Error(path,"list memory allocation error.");

	    if ((file=fopen(path,"rb"))==NULL ||
		fread(IH,(size_t)14,(size_t)1,file)!=1)
		Error(path,"image header read error.");

	    SWAP(IH,7)

	    if (IH[0]!='I'+'M'*256) Error(path,"bad magic number.");

	    l->offset=64L+(long)IH[1];

	    if (first) {
		if ((Nh=IH[2])<=0 ||
		    (Nv=IH[3])<=0) Error(path,"bad image size.");

		time1=time2=time; first=0;
	    }
	    else {
		if (Nh!=IH[2] ||
		    Nv!=IH[3]) Error(path,"image size not match.");

		if (time<time1) time1=time; else if (time>time2) time2=time;
	    }
	    l->time=time;

	    if (IH[6]!=2) Error(path,"bad image type.");

	    L=(&(l->next)); (void)fclose(file);
	}
	if (tFile!=stdin) (void)fclose(tFile);

	if (*Nt==0) Error(tPath,"empty table.");

	if (di)
{
	Table	*t=ALLOC(Table,(*Nt==1)?2:*Nt);

	if ((*T=t)==NULL) Error(tPath,tmae);

	for (*L=NULL; (l=next)!=NULL; t++) {
	    t->u.f.path=l->path;
	    t->u.f.offset=l->offset;
	    t->time=l->time; next=l->next; free(l);
	}
}
	else
{
	TableT	*tt=ALLOC(TableT,*Nt);

	if ((*T=(Table *)tt)==NULL) Error(tPath,tmae);

	for (*L=NULL; (l=next)!=NULL; tt++) {
#ifdef	OM
	    tt->u.f.path=l->path;
	    tt->u.f.offset=l->offset;
#else
	    tt->path=l->path;
#ifdef	OF
	    tt->u.offset=l->offset;
#else
	    tt->offset=l->offset;
#endif
#endif
	    tt->time=l->time; next=l->next; free(l);
	}
}}

static int	Compare(t1,t2)
Table		*t1,*t2;
{
	return (t1->time>t2->time)?1:(t1->time<t2->time)?-1:0;
}

#define SOW	sizeof(Word)
#define SPRINTF	(void)sprintf

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nd,Ni,Nt,a,v1,v2,BPS,n,v,h;
	Table	*Td,*Ti,*t;
	TableT	*Tt,*tt;
	double	dr,base,step,t0,f0,df,fC,r0,D,I,T,f1,f2;
	Cell	C,**c;
	Float	**P,**F;
	long	W,O;
	size_t	N;
	Word	*memory;
	FILE	*file;
	char	slice[STR_LEN],format[STR_LEN],str[STR_LEN],*ptr,desc[STR_LEN],
		*imae="image memory allocation error.",
		*idre="image data read error.";

	if (argc!=9 && argc!=11 && argc!=12 && argc!=14)
	Error("","usage : hp2tg {d.tbl i.tbl t.tbl} slice1 slice2\n"
		 "              dr r0/dr_base r0/dr_step t0 {f0 df} BPS TG/");

	if (argc==9 || argc==11) {
	    Check("-",&Nd,1,&Td);
	    Check("-",&Ni,1,&Ti);
	    Check("-",&Nt,0,(Table **)&Tt); a=1;
	}
	else {
	    Check(argv[1],&Nd,1,&Td);
	    Check(argv[2],&Ni,1,&Ti);
	    Check(argv[3],&Nt,0,(Table **)&Tt); a=4;
	}
	if ((v1=atoi(argv[a+0]))<0 ||
	    (v2=atoi(argv[a+1]))>=Nv || v1>v2) Error("","bad slice range.");

#define DR	argv[a+2]

	if ((dr=atof(DR))<=0.0) Error("","bad detector interval.");

	base=atof(argv[a+3]);
	step=atof(argv[a+4]);

#define T0	argv[a+5]

	t0=atof(T0)*M_PI/180.0;

	if ((BPS=atoi(argv[argc-2]))<=0 || BPS>8*sizeof(Cell))
	    Error("","bad bit per sample (BPS).");

	C=(Cell)((1<<BPS)-1);

	if (argc==11 || argc==14) {
	    if ((df=atof(argv[argc-3]))<=0.0) Error("","bad CT-value step.");

	    f0=atof(argv[argc-4]); fC=f0+df*(double)C; f0-=(df/2.0);
	}
	W=(long)Nh*SOW; O=(long)v1*W; N=(size_t)(v2-v1+1)*(size_t)Nh;

	for (t=Td, n=0; n<Nd; n++, t++) {
	    if ((memory=ALLOC(Word,N))==NULL) Error(t->u.f.path,imae);

	    if ((file=fopen(t->u.f.path,"rb"))==NULL ||
		fseek(file,t->u.f.offset+O,SEEK_SET) ||
		fread(memory,SOW,N,file)!=N) Error(t->u.f.path,idre);

	    SWAP(memory,N)

	    (void)fclose(file); free(t->u.f.path); t->u.memory=memory;
	}
	if (Nd==1) {
	    Td->time=time1; t->u.memory=memory; t->time=time2; Nd=2;
	}
	else
	    qsort(Td,(size_t)Nd,sizeof(Table),Compare);

	for (t=Ti, n=0; n<Ni; n++, t++) {
	    if ((memory=ALLOC(Word,N))==NULL) Error(t->u.f.path,imae);

	    if ((file=fopen(t->u.f.path,"rb"))==NULL ||
		fseek(file,t->u.f.offset+O,SEEK_SET) ||
		fread(memory,SOW,N,file)!=N) Error(t->u.f.path,idre);

	    SWAP(memory,N)

	    (void)fclose(file); free(t->u.f.path); t->u.memory=memory;
	}
	if (Ni==1) {
	    Ti->time=time1; t->u.memory=memory; t->time=time2; Ni=2;
	}
	else
	    qsort(Ti,(size_t)Ni,sizeof(Table),Compare);

	for (tt=Tt, n=0; n<Nt; n++, tt++) {
#ifdef	OM
	    if ((memory=ALLOC(Word,N))==NULL) Error(tt->u.f.path,imae);

	    if ((file=fopen(tt->u.f.path,"rb"))==NULL ||
		fseek(file,tt->u.f.offset+O,SEEK_SET) ||
		fread(memory,SOW,N,file)!=N) Error(tt->u.f.path,idre);

	    SWAP(memory,N)

	    (void)fclose(file); free(tt->u.f.path); tt->u.memory=memory;
#else
#ifdef	OF
	    if ((file=fopen(tt->path,"rb"))==NULL ||
		fseek(file,tt->u.offset+O,SEEK_SET)) Error(tt->path,idre);

	    tt->u.file=file;
#endif
#endif
	    for (t=Td+Nd-2; t!=Td && t->time>tt->time; t--) ;

	    tt->Md1=t->u.memory; tt->Md2=t[1].u.memory;
	    tt->Rd1=1.0-(tt->Rd2=(tt->time-t->time)/(t[1].time-t->time));

	    for (t=Ti+Ni-2; t!=Ti && t->time>tt->time; t--) ;

	    tt->Mi1=t->u.memory; tt->Mi2=t[1].u.memory;
	    tt->Ri1=1.0-(tt->Ri2=(tt->time-t->time)/(t[1].time-t->time));
	}
#ifndef	OM
	if ((memory=ALLOC(Word,Nh))==NULL) Error("",imae);
#endif
	if ((P=InitCBP(Nh,Nt))==NULL) Error("","CBP memory allocation error.");

	SPRINTF(slice,"%d",Nv-1);
	SPRINTF(format,"%%0%ud",(unsigned)strlen(slice));
#ifdef	OM
	for (N=0, v=v1; v<=v2; v++, N+=Nh) {
#else
#ifdef	OF
	for (v=v1; v<=v2; v++) {
#else
	for (v=v1; v<=v2; v++, O+=W) {
#endif
#endif
	    for (tt=Tt, n=0; n<Nt; n++, tt++) {
#ifdef	OM
		memory=tt->u.memory+N;
#else
#ifdef	OF
		if (fread(memory,SOW,(size_t)Nh,tt->u.file)!=Nh)
#else
		if ((file=fopen(tt->path,"rb"))==NULL ||
		    fseek(file,tt->offset+O,SEEK_SET) ||
		    fread(memory,SOW,(size_t)Nh,file)!=Nh)
#endif
		    Error(tt->path,idre);

		SWAP(memory,Nh)
#ifndef	OF
		(void)fclose(file);
#endif
#endif
		for (h=0; h<Nh; h++) {
		    D=tt->Rd1*(double)*(tt->Md1++)+tt->Rd2*(double)*(tt->Md2++);
		    I=tt->Ri1*(double)*(tt->Mi1++)+tt->Ri2*(double)*(tt->Mi2++);
		    T=(double)memory[h];
		    P[n][h]=(I>D && T>D)?log((I-D)/(T-D)):0.0;
		}
	    }
	    F=CBP(dr,r0=base+step*(double)v,t0);

	    f1=f2=(**F);
	    for (n=0; n<Nh; n++)
	    for (h=0; h<Nh; h++)
		if (F[n][h]<f1) f1=F[n][h]; else if (F[n][h]>f2) f2=F[n][h];

	    SPRINTF(slice,format,v);
	    (void)printf("%s\t%lf\t%lf\t%lf\n",slice,r0,f1,f2);

	    SPRINTF(str,"%21.18lf",r0*dr); ptr=str+strlen(str);
					   while (*(--ptr)=='0') *ptr='\0';
					   if (*ptr=='.') *ptr='\0';

	    if (argc==11 || argc==14)
		SPRINTF(desc,"%s\t%s\n%d\t%s\n%lf\t%lf\n%s\t%s",
			     DR,str,Nt,T0,f1,f2,argv[argc-4],argv[argc-3]);
	    else {
		SPRINTF(desc,"%s\t%s\n%d\t%s\n%lf\t%lf",DR,str,Nt,T0,f1,f2);

		df=(f2-f1)/(double)C; fC=f2; f0=f1-df/2.0;
	    }
	    c=(Cell **)F;
	    for (n=0; n<Nh; n++)
	    for (h=0; h<Nh; h++)
		c[n][h]=(F[n][h]<f0)?0:(F[n][h]>fC)?C:(Cell)((F[n][h]-f0)/df);

	    SPRINTF(str,"%s/%s.tif",argv[argc-1],slice);
	    StoreImageFile(str,Nh,Nh,BPS,c,desc);
	}
	TermCBP();
#ifdef	OM
	for (n=Nt-1; n>=0; n--) free(Tt[n].u.memory);
#else
#ifdef	OF
	for (n=Nt-1; n>=0; n--) (void)fclose(Tt[n].u.file);
#endif
	free(memory); for (n=Nt-1; n>=0; n--) free(Tt[n].path);
#endif
	for (n=(Ni==2 && Ti[1].u.memory==Ti->u.memory)?0:Ni-1; n>=0; n--)
	    free(Ti[n].u.memory);

	for (n=(Nd==2 && Td[1].u.memory==Td->u.memory)?0:Nd-1; n>=0; n--)
	    free(Td[n].u.memory);

	free(Tt); free(Ti); free(Td); return 0;
}
