
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "sif.h"

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

static FILE	*hp;

#define Get1()	fgetc(hp)

static int	Get2()
{
	int	lo,hi;

	return ((lo=Get1())==EOF || (hi=Get1())==EOF)?-1:(hi<<8)|lo;
}

static Cell	GetCell()
{
	int	cell=Get2();

	if (cell<0) Error("unexpected end of file."); return (Cell)cell;
}

#define BPS	8*sizeof(Cell)

#define MA(cnt,ptr)	malloc((cnt)*sizeof(*(ptr)))

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	len,Nx,Ny,idx;
	char	*comm=NULL;

	if (argc!=3) Error("usage : hp2tiff HiPic TIFF");

	if (argv[1][0]=='-')
	    hp=stdin;
	else
	    if ((hp=fopen(argv[1],"rb"))==NULL) Error("file not found.");

	if (Get1()!='I' || Get1()!='M') Error("bad magic number.");

	if ((len=Get2())<0) Error("bad comment length.");

	if ((Nx=Get2())<=0 ||
	    (Ny=Get2())<=0) Error("bad image size.");

	for (idx=8; idx<12; idx++) (void)Get1();	/* image offset */

	if (Get2()!=2) Error("bad image type.");

	for (idx=14; idx<64; idx++) (void)Get1();	/* reserved data */

	if (len) {
	    if ((comm=(char *)MA(len+1,comm))==NULL)
		Error("comment memory allocation error.");

	    for (idx=0; idx<len; idx++) comm[idx]=Get1(); comm[len]='\0';
	}

#ifdef	FUNCTION_CELL
	StoreImageFile(argv[2],Nx,Ny,BPS,GetCell,comm);
#else
{
	Cell	**cell;
	int	y,x;

	if ((cell=(Cell **)MA(Ny,cell))==NULL)
	    Error("cell memory allocation error (1).");

	for (y=0; y<Ny; y++) {
	    if ((cell[y]=(Cell *)MA(Nx,*cell))==NULL)
		Error("cell memory allocation error (2).");

	    for (x=0; x<Nx; x++) cell[y][x]=GetCell();
	}
	StoreImageFile(argv[2],Nx,Ny,BPS,cell,comm);
}
#endif
	if (hp!=stdin) fclose(hp); return 0;
}
