
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "cell.h"
#include "sif.h"

#define time	tick	/* to prevent error of lint */

static void	Error(path,msg)
char		*path,*msg;
{
	if (*path!='\0') (void)fprintf(stderr,"%s : ",path);

	(void)fputs(msg,stderr); (void)fputc('\n',stderr); exit(1);
}

static int	fGetW(file)
FILE		*file;
{
	int	lo,hi;

	return ((lo=fgetc(file))==EOF || (hi=fgetc(file))==EOF)?-1:(hi<<8)|lo;
}

typedef unsigned short	Word;

typedef struct {
		union {
			char	*path;
			Word	*data;
		} u;
		double	time;
	} Table;

#define Path	u.path
#define Data	u.data
#define Time	time

static char	*fnf="file not found.";
static int	Nh,Nv,first=1;
static double	time1,time2;

#define STR_LEN		1024
#define ALLOC(type,noe)	(type *)malloc(sizeof(type)*(size_t)(noe))

static void	Check(tPath,N,T,di)
char		*tPath;
int		*N,di;
Table		**T;
{
	FILE	*tFile,*iFile;
	char	line[STR_LEN],iPath[STR_LEN];
	double	time;

	struct List {
			Table		table;
			struct List	*next;
		} *list,*next,**ptr=(&next);

	Table	*table;

	if (*tPath=='-')
	    tFile=stdin;
	else
	    if ((tFile=fopen(tPath,"r"))==NULL) Error(tPath,fnf);

	for (*N=0; fgets(line,STR_LEN,tFile)!=NULL &&
		   sscanf(line,"%s %lf",iPath,&time)==2; (*N)++) {
	    if ((*ptr=list=ALLOC(struct List,1))==NULL ||
	        (list->table.Path=strdup(iPath))==NULL)
		Error(iPath,"list memory allocation error.");

	    list->table.Time=time;
	    ptr=(&(list->next));

	    if ((iFile=fopen(iPath,"rb"))==NULL) Error(iPath,fnf);

	    if (fgetc(iFile)!='I' ||
	        fgetc(iFile)!='M') Error(iPath,"bad magic number.");

	    if (fGetW(iFile)<0) Error(iPath,"bad comment length.");

	    if (first) {
		if ((Nh=fGetW(iFile))<0 ||
		    (Nv=fGetW(iFile))<0) Error(iPath,"bad image size.");

		time1=time2=time; first=0;
	    }
	    else {
		if (fGetW(iFile)!=Nh ||
		    fGetW(iFile)!=Nv) Error(iPath,"image size not match.");

		if (time<time1) time1=time; else if (time>time2) time2=time;
	    }
	    (void)fseek(iFile,4L,SEEK_CUR);

	    if (fGetW(iFile)!=2) Error(iPath,"bad image type.");

	    (void)fclose(iFile);
	}
	if (tFile!=stdin) (void)fclose(tFile);

	if (*N==0) Error(tPath,"empty table.");

	if ((*T=table=ALLOC(Table,(di && *N==1)?2:*N))==NULL)
	    Error(tPath,"table memory allocation error.");

	for (*ptr=NULL; (list=next)!=NULL; table++) {
	    *table=list->table; next=list->next; free(list);
	}
}

static int	h1,v1,h2,v2;

static Word	*Read(path,data)
char		*path;
Word		*data;
{
	int	v,h,d,n=h2-h1+1;
	FILE	*file;
	Word	*D;
	short	s=127;
	char	*c=(char *)&s,*dre="data read error.";
	long	o1=h1*sizeof(Word),o2=(Nh-1-h2)*sizeof(Word);

	if (data==NULL)
	    if ((data=ALLOC(Word,n*(v2-v1+1)))==NULL)
		Error(path,"data memory allocation error.");

	if ((file=fopen(path,"rb"))==NULL) Error(path,fnf);

	(void)fseek(file,2L,SEEK_SET);
	(void)fseek(file,(long)(60+fGetW(file)+Nh*v1*sizeof(Word)),SEEK_CUR);

	D=data;
	if (*c==127)
	    for (v=v1; v<=v2; v++, D+=n) {
		(void)fseek(file,o1,SEEK_CUR);

		if (fread(D,sizeof(Word),(size_t)n,file)!=n) Error(path,dre);

		(void)fseek(file,o2,SEEK_CUR);
	    }
	else
	    for (v=v1; v<=v2; v++) {
		(void)fseek(file,o1,SEEK_CUR);

		for (h=h1; h<=h2; h++, D++) {
		    if ((d=fGetW(file))<0) Error(path,dre);

		    *D=(Word)d;
		}
		(void)fseek(file,o2,SEEK_CUR);
	    }

	(void)fclose(file);

	free(path); return data;
}

static int	Compare(t1,t2)
Table		*t1,*t2;
{
	return (t1->Time>t2->Time)?1:(t1->Time<t2->Time)?-1:0;
}

#define SPRINTF	(void)sprintf

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nd,Ni,Nt,n,Ny,b_s,BPS,v,h;
	Table	*Td,*Ti,*Tt,*T;
	double	base,step,roof,**array,**Ay,
		time,R1d,R2d,R1i,R2i,d,i,t,a,a1,a2,*Ax;
	Cell	c,*C;
	char	desc[STR_LEN],*amae="array memory allocation error.";
	Word	*D,*D1d,*D2d,*D1i,*D2i,*data=NULL;

	if (argc!=3 && argc!=6 && argc!=8 && argc!=10 &&
	    argc!=5 && argc!=7 && argc!=9 && argc!=12)
	    Error("usage",
#if	defined(XTS)
	    "hp2xts {d.tbl i.tbl t.tbl} {h1 v1 h2 v2} {base step} BPS XT/"
#elif	defined(XPS)
	    "hp2xps {d.tbl i.tbl t.tbl} {h1 v1 h2 v2} {base step} BPS XP/"
#else
	    "hp2xta {d.tbl i.tbl t.tbl} {h1 v1 h2 v2} {base step} BPS xta.tif"
#endif
		 );

	if (argc==3 || argc==5 || argc==7 || argc==9) {
	    Check("-",&Nd,&Td,1);
	    Check("-",&Ni,&Ti,1);
	    Check("-",&Nt,&Tt,0);
	}
	else {
	    Check(argv[1],&Nd,&Td,1);
	    Check(argv[2],&Ni,&Ti,1);
	    Check(argv[3],&Nt,&Tt,0);
	}
	if (argc==3 || argc==5 || argc==6 || argc==8) {
	    h1=v1=0; h2=Nh-1; v2=Nv-1;
	}
	else {
	    n=(argc==7 || argc==9)?0:3;
	    if ((h1=atoi(argv[n+1]))<0 || (h2=atoi(argv[n+3]))>=Nh || h1>h2 ||
		(v1=atoi(argv[n+2]))<0 || (v2=atoi(argv[n+4]))>=Nv || v1>v2)
		Error("","bad image area.");
	}
	Ny=v2-v1+1;

	if ((BPS=atoi(argv[argc-2]))<=0 || BPS>8*sizeof(Cell))
	    Error("","bad bit per sample (BPS).");

	c=(Cell)((1<<BPS)-1);

	if (b_s=argc==5 || argc==8 || argc==9 || argc==12) {
	    if ((step=atof(argv[argc-3]))<=0.0) Error("","bad value step.");

	    base=atof(argv[argc-4]); roof=base+step*(double)c;
	}
	for (T=Td, n=0; n<Nd; n++, T++) T->Data=Read(T->Path,(Word *)NULL);

	if (Nd==1) {
	    Td->Time=time1; T->Time=time2; T->Data=Td->Data; Nd=2;
	}
	else
	    qsort(Td,(size_t)Nd,sizeof(Table),Compare);

	for (T=Ti, n=0; n<Ni; n++, T++) T->Data=Read(T->Path,(Word *)NULL);

	if (Ni==1) {
	    Ti->Time=time1; T->Time=time2; T->Data=Ti->Data; Ni=2;
	}
	else
	    qsort(Ti,(size_t)Ni,sizeof(Table),Compare);

#define ALLOCATE(Nx)	\
	if ((array=ALLOC(double *,Ny))==NULL) Error("",amae); \
\
	for (Ay=array, v=0; v<Ny; v++, Ay++) \
	    if ((*Ay=ALLOC(double,Nx))==NULL) Error("",amae)

#define SEARCH()	D=data=Read(Tt[n].Path,data); time=Tt[n].Time; \
\
			for (T=Td+Nd-2; T!=Td && T->Time>time; T--) ; \
			D1d=T->Data; D2d=T[1].Data; \
			R1d=1.0-(R2d=(time-T->Time)/(T[1].Time-T->Time)); \
\
			for (T=Ti+Ni-2; T!=Ti && T->Time>time; T--) ; \
			D1i=T->Data; D2i=T[1].Data; \
			R1i=1.0-(R2i=(time-T->Time)/(T[1].Time-T->Time))

#define INTERPOLATE()	d=(double)*(D1d++)*R1d+(double)*(D2d++)*R2d; \
			i=(double)*(D1i++)*R1i+(double)*(D2i++)*R2i; \
			t=(double)*(D++)

#if	defined(XTS)
#define LOG(d,i,t)	((i>d && t>d)?(t-d)/(i-d):0.0)
#else
#define LOG(d,i,t)	((i>d && t>d)?log((i-d)/(t-d)):0.0)
#endif

#define STORE(Nx,path)	\
	if (b_s) \
	    SPRINTF(desc,"%lf\t%lf\n%s\t%s",a1,a2,argv[argc-4],argv[argc-3]); \
	else { \
	    base=a1; roof=a2; step=(a2-a1)/(double)c; \
	    SPRINTF(desc,"%lf\t%lf",a1,a2); \
	} \
	for (		Ay=array , v=0; v<Ny; v++, Ay++     ) \
	for (C=(Cell *)(Ax=(*Ay)), h=0; h<Nx; h++, Ax++, C++) \
	    *C=(*Ax<=base)?0:(*Ax>=roof)?c:(Cell)floor((*Ax-base)/step+0.5); \
\
	StoreImageFile(path,Nx,Ny,BPS,(Cell **)array,desc)

#if	defined(XTS) || defined(XPS)
{
	int	Nx=h2-h1+1;
	char	frame[STR_LEN],format[STR_LEN],path[STR_LEN];

	ALLOCATE(Nx);

	SPRINTF(frame,"%d",Nt-1); SPRINTF(format,"%%0%dd",(int)strlen(frame));

	for (n=0; n<Nt; n++) {
	    SEARCH();

	    for (Ay=array, v=v1; v<=v2; v++, Ay++)
	    for (Ax=(*Ay), h=h1; h<=h2; h++, Ax++) {
		INTERPOLATE(); *Ax=a=LOG(d,i,t);

		if (v==v1 && h==h1)
		    a1=a2=a;
		else
		    if (a<a1) a1=a; else if (a>a2) a2=a;
	    }
	    SPRINTF(frame,format,n);
	    (void)printf("%s\t%lf\t%lf\n",frame,a1,a2);

	    SPRINTF(path,"%s/%s.tif",argv[argc-1],frame); STORE(Nx,path);
	}
}
#else
	ALLOCATE(Nt);

	for (n=0; n<Nt; n++) {
	    SEARCH();

	    for (Ay=array, v=v1; v<=v2; v++, Ay++) {
		a=0.0;
		for (h=h1; h<=h2; h++) {
		    INTERPOLATE(); a+=LOG(d,i,t);
		}
		(*Ay)[n]=a;

		if (n==0 && v==v1)
		    a1=a2=a;
		else
		    if (a<a1) a1=a; else if (a>a2) a2=a;
	    }
	}
	(void)printf("%lf\t%lf\n",a1,a2);

	STORE(Nt,argv[argc-1]);
#endif
	return 0;
}
