
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#ifndef	M_PI
#define M_PI	3.14159265358979323846
#endif

static void	Error(char *path,char *msg)
{
	(void)fprintf(stderr,"%s : %s\n",path,msg); exit(1);
}

static int	GetW(FILE *file)
{
	int	l,h;

	return ((l=fgetc(file))==EOF || (h=fgetc(file))==EOF)?-1:(h<<8)|l;
}

static FILE	*Open(char *path,int first,int *X,int *Y)
{
	FILE	*f;
	int	l;

	if ((f=fopen(path,"rb"))==NULL) Error(path,"file not found.");

	if (fgetc(f)!='I' || fgetc(f)!='M') Error(path,"bad magic number.");

	if ((l=GetW(f))<0) Error(path,"bad comment length.");

	if (first) {
	    if ((*X=GetW(f))<=0 ||
		(*Y=GetW(f))<=0) Error(path,"bad image size.");
	}
	else
	    if (*X!=GetW(f) ||
		*Y!=GetW(f)) Error(path,"image size not match.");

	(void)fseek(f,4L,SEEK_CUR);

	if (GetW(f)!=2) Error(path,"bad image type.");

	(void)fseek(f,50L+l,SEEK_CUR);

	return f;
}

static int	GetI(char *path,FILE *file)
{
	int	i;

	if ((i=GetW(file))<0) Error(path,"unexpected end of file.");

	return i;
}

int	main(int argc,char **argv)
{
	FILE	*D,*I,*T;
	int	X,Y,y,x,d,i,t;
	double	dr,p0,p1,p2,p,r2;

	if (argc!=5) Error("usage","hp_ca D.img I.img T.img dr");

	D=Open(argv[1],1,&X,&Y);
	I=Open(argv[2],0,&X,&Y);
	T=Open(argv[3],0,&X,&Y);

	if ((dr=atof(argv[4]))<=0.0) Error(argv[4],"bad detector interval.");

	for (y=0; y<Y; y++) {
	    p0=p1=p2=0.0;
	    for (x=0; x<X; x++) {
		d=GetI(argv[1],D);
		i=GetI(argv[2],I);
		t=GetI(argv[3],T);
		if (d<i && d<t) {
		    p=log((double)(i-d)/(double)(t-d));
		    p0+=p; p1+=(p*(double)x); p2+=(p*(double)x*(double)x);
		}
	    }
	    (void)printf("%d\t%le\t",y,p0*dr);
	    if (p0<=0.0 || (r2=4.0*(p2/p0-p1*p1/(p0*p0)))<=0.0)
		(void)puts("0\t0");
	    else
		(void)printf("%lg\t%lg\n",sqrt(r2),p0/(M_PI*r2)/dr);
	}
	(void)fclose(T); (void)fclose(I); (void)fclose(D);

	return 0;
}
