
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "srf.h"

extern void	Error(char *msg);

static void	Abort(char *head,char *tail)
{
	(void)fputs(head,stderr); (void)fputs(" : ",stderr); Error(tail);
}

static int	GetW(FILE *file)
{
	int	lo,hi;

	return ((lo=fgetc(file))==EOF || (hi=fgetc(file))==EOF)?-1:hi*256+lo;
}

static FILE	*Open(char *path,int *Nx,int *Ny)
{
	FILE	*file;
	int	cl,ny;

	if ((file=fopen(path,"rb"))==NULL) Abort(path,"file not found.");

	if (fgetc(file)!='I' ||
	    fgetc(file)!='M') Abort(path,"bad magic number.");

	if ((cl=GetW(file))<0) Abort(path,"bad comment length.");

	if ((*Nx=GetW(file))<0 ||
	    ( ny=GetW(file))<0) Abort(path,"bad image size.");

	(void)fseek(file,4L,SEEK_CUR);

	if (GetW(file)!=2) Abort(path,"bad image type.");

	(void)fseek(file,50L+cl,SEEK_CUR);

	if (Ny!=NULL) *Ny=ny;

	return file;
}

typedef unsigned short	Word;

static void	Read(char *path,int y,Word *W)
{
	int	Nx,x,w;
	FILE	*file=Open(path,&Nx,NULL);

	(void)fseek(file,2L*Nx*y,SEEK_CUR);

	for (x=0; x<Nx; x++) {
	    if ((w=GetW(file))<0) Abort(path,"data read error.");

	    *(W++)=(Word)w;
	}
        (void)fclose(file);
}

typedef struct {
		int	n;
		double	c,a;
		int	t;
	} Log;

static int	Compare(Log *l1,Log *l2)
{
	int	t=l1->t-l2->t;

	return (t)?t:(l1->t<0)?(l1->c>l2->c)?1:(l1->c<l2->c)?-1:0:
		     (l1->t>0)?(l1->a>l2->a)?1:(l1->a<l2->a)?-1:0:0;
}

#define LEN	2048

int	main(int argc,char **argv)
{
	char	*O,*D,**Q,S[LEN];
	int	N,Nx,Ny,z,n,t,i,x,y,Ni=0,Nt=0;
	Log	*L;
	FILE	*F;
	double	c,a,c1,c2,r1,r2,d,i_d,t_d,p;
	Word	*Wd,**Wi,*Wt,*W1,*W2;

	if (argc!=4) Abort("usage","hp_sg HiPic/ layer SG.bin");

	SearchRawFiles(argv[1],&O,&D,&N,&Q);

	(void)sprintf(S,"%s/%s",argv[1],D);
	(void)fclose(Open(S,&Nx,&Ny));

	if ((z=atoi(argv[2]))<0 || z>=Ny) Abort(argv[2],"bad layer.");

	if ((L=(Log *)malloc(sizeof(Log)*(N+1)))==NULL)
	    Error("no memory for log data.");

	for (n=0; n<=N; n++) L[n].t=0;

	(void)sprintf(S,"%s/%s",argv[1],O);
	if ((F=fopen(S,"r"))==NULL) Abort(S,"file not found.");

	while (fgets(S,LEN,F)!=NULL &&
	       sscanf(S,"%d %lf %lf %d%n",&n,&c,&a,&t,&i)==4) {
	    S[i]='\0';

	    if (n<0 || n>=N || t<0 || t>1) Abort(S,"bad log data.");

	    if (L[n].t!=0) Abort(S,"duplicated image number.");

	    if (Q[n]==NULL) Abort(S,"missing image file.");

	    (void)sprintf(S,"%s/%s",argv[1],Q[n]);
	    (void)fclose(Open(S,&x,&y));

	    if (Nx!=x || Ny!=y) Abort(S,"image size not match.");

	    L[n].n=n; L[n].c=c; L[n].a=a; L[n].t=t*2-1;

	    if (Ni+Nt==0) c1=c2=c; else if (c1>c) c1=c;
				   else if (c2<c) c2=c;

	    if (t==0) ++Ni; else ++Nt;
	}
	(void)fclose(F);

	if (Ni==0 || Nt==0) Error("no I0- and/or I-image file.");

	if ((Wd=(Word *)malloc((size_t)2*Nx))==NULL ||
	    (Wi=(Word **)malloc(sizeof(Word *)*Ni))==NULL ||
	    (*Wi=(Word *)malloc((size_t)Ni*2*Nx))==NULL ||
	    (Wt=(Word *)malloc((size_t)2*Nx))==NULL)
	    Error("no memory for image data.");

	(void)sprintf(S,"%s/%s",argv[1],D);
	Read(S,z,Wd);

	qsort(L,N+1,sizeof(Log),(int (*)())Compare);

	for (n=0; n<Ni; n++) {
	    if (n>0) Wi[n]=Wi[n-1]+(size_t)Nx;

	    (void)sprintf(S,"%s/%s",argv[1],Q[L[n].n]);
	    Read(S,z,Wi[n]);
	}
	if (Ni==1) {
	    Ni=2; L[1].n=L[0].n; L[0].c=c1; L[1].c=c2;
	}
	if (argv[3][0]=='-')
	    F=stdout;
	else
	    if ((F=fopen(argv[3],"wb"))==NULL) Abort(argv[3],"file not open.");

	(void)fprintf(F,"%d\t%d\n",Nx,Nt);

	for (n=N+1-Nt; n<=N; n++) {
	    (void)sprintf(S,"%s/%s",argv[1],Q[L[n].n]);
	    Read(S,z,Wt);

	    c=L[n].c; for(i=Ni-2; i>0; i--) if (c>=L[i].c) break;

	    W1=Wi[i]; W2=Wi[i+1]; r1=1.0-(r2=(c-L[i].c)/(L[i+1].c-L[i].c));

	    for (x=0; x<Nx; x++) {
		d=(double)Wd[x];
		p=((i_d=r1*(double)W1[x]+r2*(double)W2[x]-d)>0.0 &&
		   (t_d=(double)Wt[x]-d)>0.0)?log(i_d/t_d):0.0;
		(void)fwrite(&p,sizeof(p),(size_t)1,F);
	    }
	}
	if (F!=stdout) (void)fclose(F);

	free(Wt); free(*Wi); free(Wi); free(Wd); free(L);

	for (n=N-1; n>=0; n--) if (Q[n]!=NULL) free(Q[n]);

	free(Q); free(D); free(O);

	return 0;
}
