
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif_rgb.h"
#include "sif.h"
#include "sif_rgb.h"

extern void	Error();

#define ALLOC(type,noe)	(type *)malloc(sizeof(type)*(size_t)(noe))

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nx,Ny,Sx,Sy,r,g,b,x,y,a,Ph,Ch,Pv,Cv,v,h,gray,BPS,V,H;
	char	*tmae="table memory allocation error.",
		*imae="image memory allocation error.";
	Cell	**Pr,**Pg,**Pb,**Cr,**Cg,**Cb;
	double	ratio;

	struct Table {
		char	*path;
		int	H,V;
	} **T;

	if (argc<10)
	    Error("usage : pile_RGB Nx Ny Sx Sy R G B parent children");

#define NX	(argv[1])
#define NY	(argv[2])
#define SX	(argv[3])
#define SY	(argv[4])

	if ((Nx=abs(atoi(NX)))==0 ||
	    (Ny=abs(atoi(NY)))==0) Error("bad number of images.");

	if (argc-9!=Nx*Ny)
	    (void)fputs("number of images not match (warning).\n",stderr);

	Sx=abs(atoi(SX));
	Sy=abs(atoi(SY));

	if ((r=atoi(argv[5]))<0 || r>255 ||
	    (g=atoi(argv[6]))<0 || g>255 ||
	    (b=atoi(argv[7]))<0 || b>255)
	    Error("bad R, G and/or B value for space filling.");

	if ((T=ALLOC(struct Table *,Nx))==NULL) Error(tmae);

	for (x=0; x<Nx; x++) {
	    if ((T[x]=ALLOC(struct Table,Ny))==NULL) Error(tmae);

	    for (y=0; y<Ny; y++) {
		T[x][y].path=NULL; T[x][y].H=T[x][y].V=0;
	    }
	}

#define CHECK	{ \
	if (argv[a][0]!='-') \
	    ReadImageFile_RGB(T[x][y].path=argv[a],&(T[x][y].H),&(T[x][y].V), \
			      NULL,NULL,NULL,NULL,NULL); \
}

	a=9;
	if (*NX!='-')
	    for (x=0; x<Nx && a<argc; x++)
		if (*NY!='-')
		    for (y=0; y<Ny && a<argc; y++, a++) CHECK
		else
		    for (y=Ny-1; y>=0 && a<argc; y--, a++) CHECK
	else
	    if (*NY!='-')
		for (y=0; y<Ny && a<argc; y++)
		    for (x=0; x<Nx && a<argc; x++, a++) CHECK
	    else
		for (y=Ny-1; y>=0 && a<argc; y--)
		    for (x=0; x<Nx && a<argc; x++, a++) CHECK

	for (Ph=(*SX!='-')?Sx:0, x=0; x<Nx; x++, Ph+=(Ch+Sx)) {
	    Ch=0; for (y=0; y<Ny; y++) if (T[x][y].H>Ch) Ch=T[x][y].H;

	    for (y=0; y<Ny; y++) T[x][y].H=Ph+(Ch-T[x][y].H)/2;
	}
	if (*SX=='-') Ph-=Sx;

	for (Pv=(*SY!='-')?Sy:0, y=0; y<Ny; y++, Pv+=(Cv+Sy)) {
	    Cv=0; for (x=0; x<Nx; x++) if (T[x][y].V>Cv) Cv=T[x][y].V;

	    for (x=0; x<Nx; x++) T[x][y].V=Pv+(Cv-T[x][y].V)/2;
	}
	if (*SY=='-') Pv-=Sy;

	if ((Pr=ALLOC(Cell *,Pv))==NULL ||
	    (Pg=ALLOC(Cell *,Pv))==NULL ||
	    (Pb=ALLOC(Cell *,Pv))==NULL) Error(imae);

	for (v=0; v<Pv; v++) {
	    if ((Pr[v]=ALLOC(Cell,Ph))==NULL ||
		(Pg[v]=ALLOC(Cell,Ph))==NULL ||
		(Pb[v]=ALLOC(Cell,Ph))==NULL) Error(imae);

	    for (h=0; h<Ph; h++) {
		Pr[v][h]=(Cell)r; Pg[v][h]=(Cell)g; Pb[v][h]=(Cell)b;
	    }
	}
	gray=r==g && r==b;

	for (x=0; x<Nx; x++)
	for (y=0; y<Ny; y++)
	    if (T[x][y].path!=NULL) {
		ReadImageFile_RGB(T[x][y].path,&Ch,&Cv,&BPS,&Cr,&Cg,&Cb,NULL);

		ratio=255.0/(double)((1<<BPS)-1);
		for (V=T[x][y].V, v=0; v<Cv; v++, V++) {
		    for (H=T[x][y].H, h=0; h<Ch; h++, H++) {
			Pr[V][H]=(Cell)(ratio*(double)Cr[v][h]+0.5);
			Pg[V][H]=(Cell)(ratio*(double)Cg[v][h]+0.5);
			Pb[V][H]=(Cell)(ratio*(double)Cb[v][h]+0.5);
			gray&=(Cr[v][h]==Cg[v][h] && Cr[v][h]==Cb[v][h]);
		    }
		    free(Cb[v]); free(Cg[v]); free(Cr[v]);
		}
		free(Cb); free(Cg); free(Cr);
	    }

	if (gray)
	    StoreImageFile(argv[8],Ph,Pv,8,Pr,NULL);
	else
	    StoreImageFile_RGB(argv[8],Ph,Pv,8,Pr,Pg,Pb,NULL);

	return 0;
}
