
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cell.h"
#include "rif_rgb.h"
#include "rif.h"
#include "sif_rgb.h"
#include "sif.h"

extern void	Error();

static void	ReadTIFF(TIFF,Nx,Ny,R,G,B)
char		*TIFF;
int		*Nx,*Ny;
Cell		***R,***G,***B;
{
	int	BPS,y,x;
	double	ratio;

	ReadImageFile_RGB(TIFF,Nx,Ny,&BPS,R,G,B,NULL);

	ratio=255.0/(double)((1<<BPS)-1);
	for (y=0; y<*Ny; y++)
	for (x=0; x<*Nx; x++) {
	    (*R)[y][x]=(Cell)((double)(*R)[y][x]*ratio+0.5);
	    (*G)[y][x]=(Cell)((double)(*G)[y][x]*ratio+0.5);
	    (*B)[y][x]=(Cell)((double)(*B)[y][x]*ratio+0.5);
	}
}

#define GS_LEN	2048
#define FPRINTF	(void)fprintf

static void	PutLabel(TIFF,Nx,Ny,R,G,B,GS,name,size,r,g,b,h,v,label)
char		*TIFF,*GS,*name,*label;
int		Nx,Ny,r,g,b,h,v;
Cell		**R,**G,**B;
double		size;
{
	char	gs[GS_LEN],*ptr;
	FILE	*file;
	Cell	**mask;
	int	y,x;

	(void)sprintf(gs,
		      "%s -q -r72 -g%dx%d -sDEVICE=tifflzw -sOutputFile=%s -",
		      GS,Nx,Ny,TIFF);
	if ((file=popen(gs,"w"))==NULL) Error("gs initialization error.");

	FPRINTF(file,"/%s findfont %lf scalefont setfont\n"
		     "0 [\n(",name,size);
	for (ptr=label; *ptr!='\0'; ptr++)
	    switch(*ptr) {
		case '\\' :
		    if (ptr[1]=='n') {
			(void)fputs(")\n(",file); ++ptr; break;
		    }
		case '(' :
		case ')' : (void)fputc('\\',file);
		default : (void)fputc((int)*ptr,file);
	    }

	(void)fputs(")\n] {\n",file);
	switch(h) {
	    case 'L' :
		FPRINTF(file,"%lf 0.5 mul\n",size); break;
	    case 'C' :
		FPRINTF(file,"dup stringwidth pop %d sub 2 div neg\n",
			     Nx); break;
	    case 'R' :
		FPRINTF(file,"dup stringwidth pop %d sub %lf 0.5 mul add neg\n",
			     Nx,size); break;
	}
	FPRINTF(file,"3 -1 roll dup %lf mul\n",size);
	switch(v) {
	    case 'U' :
		FPRINTF(file,"%d %lf 1.375 mul sub exch sub\n",Ny,size); break;
	    case 'D' :
		FPRINTF(file,"%lf 0.625 mul add\n",size); break;
	}
	(void)fputs("exch 1 add 4 1 roll\n"
		    "newpath moveto show\n"
		    "} forall pop showpage\n",file);

	if (pclose(file)==(-1)) Error("gs termination error.");

	ReadImageFile(TIFF,NULL,NULL,NULL,&mask,NULL);

	for (y=0; y<Ny; y++) {
	    for (x=0; x<Nx; x++)
		if (mask[y][x]) {
		    R[y][x]=(Cell)r; G[y][x]=(Cell)g; B[y][x]=(Cell)b;
		}

	    free(mask[y]);
	}
	free(mask);
}

static void	StoreTIFF(TIFF,Nx,Ny,R,G,B)
char		*TIFF;
int		Nx,Ny;
Cell		**R,**G,**B;
{
	int	y,x;

	for (y=0; y<Ny; y++)
	for (x=0; x<Nx; x++)
	    if (R[y][x]!=G[y][x] || G[y][x]!=B[y][x]) {
		StoreImageFile_RGB(TIFF,Nx,Ny,8,R,G,B,NULL);

		goto Break;
	    }

	StoreImageFile(TIFF,Nx,Ny,8,R,NULL);
Break:;
	for (y=0; y<Ny; y++) {
	    free(B[y]); free(G[y]); free(R[y]); 
	}
	free(B); free(G); free(R);
}

static double	GetSize(Size)
char		*Size;
{
	double	size=atof(Size);

	return (size>=0.0)?size:0.0;
}

static int	GetRGB(RGB)
char		*RGB;
{
	int	rgb=atoi(RGB);

	return (rgb>=0 && rgb<=255)?rgb:256;
}

static int	GetHV(HV,h,v)
char		*HV;
int		*h,*v;
{
	return ((*h=(int)HV[0]&223)=='L' || *h=='C' || *h=='R') &&
	       ((*v=(int)HV[1]&223)=='U' || *v=='D');
}

#define STR_LEN	4096
#define DELIM	" \t\n"

int	main(argc,argv)
int	argc;
char	**argv;
{
	char	*GS,desc[STR_LEN],*name,str[STR_LEN],*tok,label[STR_LEN],*TIFF,
		*gs="gs",
		*bd="bad description (skipping) :\n%s",
		*bfs="bad font size.",
		*bfRGB="bad font R, G and/or B.",
		*blp="bad label position ([L,C,R][U,D]).";
	int	Nx,Ny,r,g,b,h,v,a;
	double	size;
	Cell	**R,**G,**B;

	if (argc!=2 && argc!=7 && argc<8)
	    Error(
	    "usage : put_label TIFF\n"
	    "        put_label      fontName size R G B [L,C,R][U,D]\n"
	    "        put_label TIFF fontName size R G B [L,C,R][U,D] label"
	    );

	if ((GS=getenv("GS"))==NULL) GS=gs;

switch(argc) {
    case 2 :
	ReadTIFF(argv[1],&Nx,&Ny,&R,&G,&B);

	while (fgets(desc,STR_LEN,stdin)!=NULL)
	    if ((name=strtok(strcpy(str,desc),DELIM))!=NULL)
	    if ((tok=strtok(NULL,DELIM))==NULL || (size=GetSize(tok))<=0.0 ||
		(tok=strtok(NULL,DELIM))==NULL || (r=GetRGB(tok))==256 ||
		(tok=strtok(NULL,DELIM))==NULL || (g=GetRGB(tok))==256 ||
		(tok=strtok(NULL,DELIM))==NULL || (b=GetRGB(tok))==256 ||
		(tok=strtok(NULL,DELIM))==NULL || !GetHV(tok,&h,&v) ||
		(tok=strtok(NULL,DELIM))==NULL)
		FPRINTF(stderr,bd,desc);
	    else {
		(void)strcpy(label,tok);
		while ((tok=strtok(NULL,DELIM))!=NULL)
		    (void)strcat(strcat(label," "),tok);

		PutLabel(argv[1],Nx,Ny,R,G,B,GS,name,size,r,g,b,h,v,label);
	    }

	StoreTIFF(argv[1],Nx,Ny,R,G,B);

	break;
    case 7 :
	if ((size=GetSize(argv[2]))<=0.0) Error(bfs);

	if ((r=GetRGB(argv[3]))==256 ||
	    (g=GetRGB(argv[4]))==256 ||
	    (b=GetRGB(argv[5]))==256) Error(bfRGB);

	if (!GetHV(argv[6],&h,&v)) Error(blp);

	while (fgets(desc,STR_LEN,stdin)!=NULL)
	    if ((TIFF=strtok(strcpy(str,desc),DELIM))!=NULL)
	    if ((tok=strtok(NULL,DELIM))==NULL)
		FPRINTF(stderr,bd,desc);
	    else {
		(void)strcpy(label,tok);
		while ((tok=strtok(NULL,DELIM))!=NULL)
		    (void)strcat(strcat(label," "),tok);

		ReadTIFF(TIFF,&Nx,&Ny,&R,&G,&B);
		PutLabel(TIFF,Nx,Ny,R,G,B,GS,argv[1],size,r,g,b,h,v,label);
		StoreTIFF(TIFF,Nx,Ny,R,G,B);
	    }

	break;
    default :
	if ((size=GetSize(argv[3]))<=0.0) Error(bfs);

	if ((r=GetRGB(argv[4]))==256 ||
	    (g=GetRGB(argv[5]))==256 ||
	    (b=GetRGB(argv[6]))==256) Error(bfRGB);

	if (!GetHV(argv[7],&h,&v)) Error(blp);

	(void)strcpy(label,argv[8]);
	for (a=9; a<argc; a++) (void)strcat(strcat(label," "),argv[a]);

	ReadTIFF(argv[1],&Nx,&Ny,&R,&G,&B);
	PutLabel(argv[1],Nx,Ny,R,G,B,GS,argv[2],size,r,g,b,h,v,label);
	StoreTIFF(argv[1],Nx,Ny,R,G,B);

	break;
}
	return 0;
}
