
#include <stdio.h>
#include <stdlib.h>
#include "cbp.h"
#include "cell.h"
#include "sif.h"

extern void	Error();

#define SPRINTF	(void)sprintf

static void	sPrintF(ptr,d)
char		*ptr;
double		d;
{
	SPRINTF(ptr,"%21.18lf",d); while (*ptr!='\0') ++ptr;
				   while (*(--ptr)=='0') *ptr='\0';
				   if (*ptr=='.') *ptr='\0';
}

#define STR_LEN	256

int	main(argc,argv)
int	argc;
char	**argv;
{
	FILE	*file;
	char	str[STR_LEN],DO[STR_LEN],*ptr,path[STR_LEN],
		*sdre="sinogram data read error.";
	int	N,M,cnt,BPS,idx,m,n,v,h;
	double	dr,base,step,r0,t0,f0,df,f3,p,f1,f2;
	Cell	c,**cell;
	Float	**P,**f;

	if (argc!=9 && argc!=11)
	    Error(
"usage : sg2tgs sinogram dr r0/dr_base r0/dr_step r0/dr_cnt t0 {f0 df} BPS TG/"
	    );

	if (argv[1][0]=='-')
	    file=stdin;
	else
	    if ((file=fopen(argv[1],"rb"))==NULL)
		Error("sinogram file not found.");

	if (fgets(str,STR_LEN,file)==NULL ||
	    sscanf(str,"%d %d",&N,&M)!=2 ||
	    N<=0 || M<=0) Error("bad number of sinogram data.");

	if ((dr=atof(argv[2]))<=0.0) Error("bad detector interval.");

	step=atof(argv[4]);
	if ((cnt=atoi(argv[5]))<=0 ||
	    (base=atof(argv[3]))>=0.0 || base<=(double)(1-N) ||
	    (r0=base+step*(double)(cnt-1))>=0.0 || r0<=(double)(1-N))
	    Error("bad detector offset (unit in detector interval).");

	t0=atof(argv[6])*M_PI/180.0;

	if ((BPS=atoi(argv[argc-2]))<=0 || BPS>8*sizeof(Cell))
	    Error("bad bit per sample (BPS).");

	c=(Cell)((1<<BPS)-1);

	if (argc==11) {
	    f0=atof(argv[7]);
	    if ((df=atof(argv[8]))<=0.0) Error("bad CT-value step.");

	    f3=f0+df*(double)c; f0-=(df/2.0);
	}

	if ((P=InitCBP(N,M))==NULL) Error("memory allocation error.");

	if (sizeof(double)!=sizeof(Float))
	    for (m=0; m<M; m++)
	    for (n=0; n<N; n++) {
		if (fread(&p,sizeof(p),(size_t)1,file)!=1) Error(sdre);

		P[m][n]=p;
	    }
	else
	    if (fread(*P,sizeof(**P),(size_t)(M*N),file)!=M*N) Error(sdre);

	if (file!=stdin) (void)fclose(file);

	for (idx=0; idx<cnt; idx++) {
	    f=CBP(dr,r0=base+step*(double)idx,t0);

	    f1=f2=(**f);
	    for (v=0; v<N; v++)
	    for (h=0; h<N; h++)
		if (f[v][h]<f1) f1=f[v][h]; else if (f[v][h]>f2) f2=f[v][h];

	    sPrintF(DO,-r0);
	    (void)printf("-%s\t%lf\t%lf\n",DO,f1,f2);

	    for (ptr=DO; *ptr!='\0'; ptr++) if (*ptr=='.') *ptr='_';
	    SPRINTF(path,"%s/%s.tif",argv[argc-1],DO);

	    sPrintF(DO,r0*dr);
	    if (argc==11)
		SPRINTF(str,"%s\t%s\n%d\t%s\n%lf\t%lf\n%s\t%s",
			    argv[2],DO,M,argv[6],f1,f2,argv[7],argv[8]);
	    else {
		SPRINTF(str,"%s\t%s\n%d\t%s\n%lf\t%lf",
			    argv[2],DO,M,argv[6],f1,f2);

		df=(f2-f1)/(double)c; f3=f2; f0=f1-df/2.0;
	    }
	    cell=(Cell **)f;
	    for (v=0; v<N; v++)
	    for (h=0; h<N; h++)
	    cell[v][h]=(f[v][h]<f0)?0:(f[v][h]>f3)?c:(Cell)((f[v][h]-f0)/df);

	    StoreImageFile(path,N,N,BPS,cell,str);
	}
	TermCBP(); return 0;
}
