
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#ifndef	Filter
#define Filter	Chesler
#else
extern double	Filter(/* int i */);
#endif

extern void	Error();

#ifndef	M_PI
#define M_PI	3.14159265358979323846
#endif

double	Ramachandran(i)
int	i;
{
	return (i==0)?1.0/4.0:(i&1)?-1.0/(M_PI*M_PI*(double)i*(double)i):0.0;
}

double	Shepp(i)
int	i;
{
	return 2.0/(M_PI*M_PI*(1.0-4.0*(double)i*(double)i));
}

double	Chesler(i)
int	i;
{
	if (i==0)
	    return 1.0/8.0-1.0/(2.0*M_PI*M_PI);
	else if (i==1 || i==(-1))
	    return 1.0/16.0-1.0/(2.0*M_PI*M_PI);
	else if (i&1)
	    return -1.0/(2.0*M_PI*M_PI*(double)i*(double)i);
{
	double	i2=(double)i*(double)i,i21=i2-1.0;

	return -(i2+1.0)/(2.0*M_PI*M_PI*i21*i21);
}}

typedef struct {
		double	x,y;
	} Vector;

static void	FFT(L,E,Z,sign)
int		L,sign;
Vector		*E,*Z;
{
	int	j,k,m,n,p,q,L2=L>>1;
	Vector	z,*e,*a,*b;
	double	x,y,w;

	j=L-1;
	for (k=L-2; k>0; k--) {
	    for (m=L2; (j^=m)&m; m>>=1) ;

	    if (j>k) {
		z=Z[j]; Z[j]=Z[k]; Z[k]=z;
	    }
	}
	for (j=2, k=1, m=L2, n=L; n>1; n=m, m>>=1, k=j, j<<=1)
	    for (e=E, p=0; p<k; p++, e+=n) {
		x=e->x; y=e->y*(double)sign;
		for (a=Z+p, q=0; q<m; q++, a+=j) {
		    z=(*(b=a+k)); b->x=a->x-(w=z.x*x-z.y*y); a->x+=w;
				  b->y=a->y-(w=z.x*y+z.y*x); a->y+=w;
		}
	    }
}

#define STR_LEN	256

int	main(argc,argv)
int	argc;
char	**argv;
{
	FILE	*P,*Q;
	char	str[STR_LEN];
	int	N,M,L2,L,n,m,l;
	double	dr,a,x,y;
	Vector	*E,*Z,*G;

	if (argc!=4)
	    Error("usage : sg_p2fp SG_projection dr SG_filtered_projection");

	if (argv[1][0]=='-')
	    P=stdin;
	else
	    if ((P=fopen(argv[1],"rb"))==NULL)
		Error("sinogram file not found.");

	if (fgets(str,STR_LEN,P)==NULL ||
	    sscanf(str,"%d %d",&N,&M)!=2 ||
	    N<=0 || M<=0) Error("bad number of sinogram data.");

	if ((dr=atof(argv[2]))<=0.0) Error("bad detector interval.");

	L=1<<(int)ceil(log((double)N)/log(2.0)); L2=L<<1;

	if ((E=(Vector *)malloc(sizeof(Vector)*(size_t)L2))==NULL ||
	    (Z=(Vector *)malloc(sizeof(Vector)*(size_t)L2))==NULL ||
	    (G=(Vector *)malloc(sizeof(Vector)*(size_t)L2))==NULL)
	    Error("memory allocation error.");

	for (n=0; n<L2; n++) {
	    a=M_PI*(double)n/(double)L2; E[n].x=cos(a); E[n].y=sin(a);

	    Z[n].x=Filter(n-L)/dr/(double)L2; Z[n].y=0.0;
	}
	FFT(L2,E,Z,-1);

	for (n=0; n<L2; n++) G[n]=Z[n];

	if (argv[3][0]=='-')
	    Q=stdout;
	else
	    if ((Q=fopen(argv[3],"wb"))==NULL)
		Error("sinogram file not stored.");

	(void)fprintf(Q,"%d\t%d\n",N,M);

	for (m=0; m<M; m++) {
	    for (l=L, n=0; n<N; n++, l++) {
		if (fread(&(Z[l].x),sizeof(double),(size_t)1,P)!=1)
		    Error("sinogram data read error.");

		Z[n].x=Z[n].y=Z[l].y=0.0;
	    }
	    for (; n<L; n++, l++) Z[n].x=Z[n].y=Z[l].x=Z[l].y=0.0;

	    FFT(L2,E,Z,-1);

	    for (n=0; n<L2; n++) {
		x=Z[n].x;
		y=Z[n].y; Z[n].x=x*G[n].x-y*G[n].y;
			  Z[n].y=x*G[n].y+y*G[n].x;
	    }
	    FFT(L2,E,Z, 1);

	    for (n=0; n<N; n++)
		(void)fwrite(&(Z[n].x),sizeof(double),(size_t)1,Q);
	}
	free(G);
	free(Z);
	free(E); if (Q!=stdout) (void)fclose(Q);
		 if (P!=stdin ) (void)fclose(P); return 0;
}
