
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include "cbp.h"
#include "cell.h"
#include "sif.h"

extern void	Error();

#define STR_LEN	256
#define FPRINTF	(void)fprintf
#define SPRINTF	(void)sprintf

int	main(argc,argv)
int	argc;
char	**argv;
{
	DIR		*dir;
	unsigned	start=0,leap=1,total=0,count=0;
	char		**arg,path[STR_LEN],str[STR_LEN],r0dr[STR_LEN],*ptr,
			*sdre="%s : sinogram data read error.\n";
	double		dr,r0,base,step,t0,f0,df,f3,p,f1,f2;
	int		BPS,slice,n,m,N,M,v,h,size=0;
	Cell		C,**cell;
	struct dirent	*ent;
	FILE		*file;
	Float		**P,**f;

	if (argc!=8 && argc!=10 && argc!=11 && argc!=13)
	Error("usage : sgs2tgs SG/ {SG_start SG_leap SG_total}\n"
	      "                dr r0/dr_base r0/dr_step t0 {f0 df} BPS TG/");

	if ((dir=opendir(argv[1]))==NULL) {
	    (void)fputs(argv[1],stderr); Error(" : directory not open.");
	}
	if (argc==11 || argc==13) {
	    if (sscanf(argv[2],"%u",&start)!=1)
		Error("bad start number of sinograms.");

	    if (sscanf(argv[3],"%u",&leap)!=1 || leap==0)
		Error("bad leap number of sinograms.");

	    if (sscanf(argv[4],"%u",&total)!=1)
		Error("bad total number of sinograms.");

	    arg=argv+5;
	}
	else
	    arg=argv+2;

#define DR	arg[0]
#define BASE	arg[1]
#define STEP	arg[2]
#define T0	arg[3]

	if ((dr=atof(DR))<=0.0) Error("bad detector interval.");

	base=atof(BASE);
	step=atof(STEP);
	t0=atof(T0)*M_PI/180.0;

	if ((BPS=atoi(argv[argc-2]))<=0 || BPS>8*sizeof(Cell))
	    Error("bad bit per sample (BPS).");

	C=(Cell)((1<<BPS)-1);

	if (argc==10 || argc==13) {
	    if ((df=atof(argv[argc-3]))<=0.0) Error("bad CT-value step.");

	    f0=atof(argv[argc-4]); f3=f0+df*(double)C; f0-=(df/2.0);
	}
	while ((ent=readdir(dir))!=NULL) {

#define EDN	ent->d_name

	    if (!strcmp(EDN,".") || !strcmp(EDN,"..")) continue;

#define CONTINUE	goto Continue

	    if (count<start || (count-start)%leap!=0) CONTINUE;

	    if (sscanf(EDN,"%d",&slice)!=1) {
		FPRINTF(stderr,"%s : bad sinogram name.\n",EDN); CONTINUE;
	    }
	    SPRINTF(path,"%s/%s",argv[1],EDN);
	    if ((file=fopen(path,"rb"))==NULL) {
		FPRINTF(stderr,"%s : file not found.\n",path); CONTINUE;
	    }

#define FCLOSE_CONTINUE	(void)fclose(file); CONTINUE

	    if (fgets(str,STR_LEN,file)==NULL ||
		sscanf(str,"%d %d",&n,&m)!=2 ||
		n<=0 || m<=0) {
		FPRINTF(stderr,"%s : bad sinogram size.\n",path);
		FCLOSE_CONTINUE;
	    }
	    if (!size) {
		if ((P=InitCBP(N=n,M=m))==NULL)
		    Error("CBP memory allocation error.");

		size=1;
	    }
	    else
		if (N!=n || M!=m) {
		    FPRINTF(stderr,"%s : sinogram size not match.\n",path);
		    FCLOSE_CONTINUE;
		}

	    if ((r0=base+step*(double)slice)>=0.0 || r0<=(double)(1-N)) {
		FPRINTF(stderr,"%lf for %s : bad detector offset.\n",r0,EDN);
		FCLOSE_CONTINUE;
	    }
	    if (sizeof(double)!=sizeof(Float))
		for (m=0; m<M; m++)
		for (n=0; n<N; n++) {
		    if (fread(&p,sizeof(p),(size_t)1,file)!=1) {
			FPRINTF(stderr,sdre,path); FCLOSE_CONTINUE;
		    }
		    P[m][n]=p;
		}
	    else
		if (fread(*P,sizeof(**P),(size_t)(M*N),file)!=M*N) {
		    FPRINTF(stderr,sdre,path); FCLOSE_CONTINUE;
		}

	    (void)fclose(file);

	    f=CBP(dr,r0,t0);

	    f1=f2=(**f);
	    for (v=0; v<N; v++)
	    for (h=0; h<N; h++)
		if (f[v][h]<f1) f1=f[v][h]; else if (f[v][h]>f2) f2=f[v][h];

	    (void)printf("%s\t%lf\t%lf\t%lf\n",EDN,r0,f1,f2);

	    SPRINTF(r0dr,"%21.18lf",r0*dr); for (ptr=r0dr; *ptr!='\0'; ptr++) ;
					    while (*(--ptr)=='0') *ptr='\0';
					    if (*ptr=='.') *ptr='\0';
	    if (argc==10 || argc==13)
		SPRINTF(str,"%s\t%s\n%d\t%s\n%lf\t%lf\n%s\t%s",
			    DR,r0dr,M,T0,f1,f2,argv[argc-4],argv[argc-3]);
	    else {
		SPRINTF(str,"%s\t%s\n%d\t%s\n%lf\t%lf",DR,r0dr,M,T0,f1,f2);

		df=(f2-f1)/(double)C; f3=f2; f0=f1-df/2.0;
	    }
	    cell=(Cell **)f;
	    for (v=0; v<N; v++)
	    for (h=0; h<N; h++)
		cell[v][h]=(f[v][h]<f0)?0:
			   (f[v][h]>f3)?C:(Cell)((f[v][h]-f0)/df);

	    SPRINTF(path,"%s/%s.tif",argv[argc-1],EDN);
	    StoreImageFile(path,N,N,BPS,cell,str);

	    if (--total==0) break;
Continue:;
	    ++count;
	}
	return 0;
}
