
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "rif.h"
#include "t_m.h"

extern void	Error();

#ifndef	M_PI
#define M_PI	3.14159265358979323846
#endif

static double	f0,df,R,h0,v0,*C,*S,**P;
static int	N,M;

static void	Store(h1,v1,h2,v2,h3,v3,c)
int		h1,v1,h2,v2,h3,v3;
long		c;
{
	int	m,n2,n;
	double	r1,s1,r2,s2,r3,s3,r,s,t31,t,*p,
		x1=(double)h1-h0,Y1=v0-(double)v1,
		x2=(double)h2-h0,y2=v0-(double)v2,
		x3=(double)h3-h0,y3=v0-(double)v3,f=f0+df*(double)c;

	for (m=0; m<M; m++) {
	    r1=R+C[m]*x1+S[m]*Y1; s1=C[m]*Y1-S[m]*x1;
	    r2=R+C[m]*x2+S[m]*y2; s2=C[m]*y2-S[m]*x2;
	    r3=R+C[m]*x3+S[m]*y3; s3=C[m]*y3-S[m]*x3;
	    if (r1>r2)
		if (r3>r1) {
		    r=r1; r1=r2; r2=r;
		    s=s1; s1=s2; s2=s;
		}
		else if (r3>r2) {
		    r=r1; r1=r2; r2=r3; r3=r;
		    s=s1; s1=s2; s2=s3; s3=s;
		}
		else {
		    r=r3; r3=r1; r1=r;
		    s=s3; s3=s1; s1=s;
		}
	    else
		if (r1>r3) {
		    r=r3; r3=r2; r2=r1; r1=r;
		    s=s3; s3=s2; s2=s1; s1=s;
		}
		else if (r2>r3) {
		    r=r2; r2=r3; r3=r;
		    s=s2; s2=s3; s3=s;
		}

	    if (r1>=(double)N || r3<0.0) continue;

	    n2=(int)floor(r2); t31=(s3-s1)/(r3-r1); p=P[m];

	    if (n2>=0) {
		r=r2-r1;
		if ((n=(int)floor(r1))==n2)
		    p[n2]+=(f*fabs(s2-s1-t31*r)*r);
		else {
		    t=f*fabs((s2-s1)/r-t31);
		    if (n<0) {
			s1=t*(-r1); n=0;
		    }
		    else {
			r1=(double)(n+1)-r1; s1=t*r1; p[n++]+=(s1*r1);
		    }
		    while (n<n2 && n<N) {
			s=s1; s1+=t; p[n++]+=(s+s1);
		    }
		    if (n<N) p[n2]+=((s1+t*r)*(r2-(double)n2));
		}
	    }
	    if (n2<N) {
		r=r3-r2;
		if ((n=(int)floor(r3))==n2)
		    p[n2]+=(f*fabs(t31*r-s3+s2)*r);
		else {
		    t=f*fabs(t31-(s3-s2)/r);
		    if (n>=N) {
			s3=t*(r3-(double)N); n=N-1;
		    }
		    else {
			r3-=(double)n; s3=t*r3; p[n--]+=(s3*r3);
		    }
		    while (n>n2 && n>=0) {
			s=s3; s3+=t; p[n--]+=(s+s3);
		    }
		    if (n>=0) p[n2]+=((s3+t*r)*((double)(n2+1)-r2));
		}
	    }
	}
}

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nh,Nv,BPS,y,x;
	Cell	**cell;
	double	dr,t,p1,p2;
	char	**back;
	long	**fore;
	FILE	*file;

	if (argc!=6 && argc!=7)
	    Error("usage : tg2sg tomogram f1 f2 dr {M} sinogram");

	ReadImageFile(argv[1],&Nh,&Nv,&BPS,&cell,NULL);

	if (Nh!=Nv)
	    (void)fputs("bad image size as a tomogram (warning).\n",stderr);

	dr=0.5*atof(argv[4]);
	f0=dr*atof(argv[2]);
	df=(dr*atof(argv[3])-f0)/(double)((1<<BPS)-1);

	N=(Nh<Nv)?Nh:Nv; R=(double)N/2.0; h0=(double)Nh/2.0; v0=(double)Nv/2.0;

	if (argc==6)
	    M=(int)ceil(M_PI*R);
	else
	    if ((M=atoi(argv[5]))<=0) Error("bad number of projections.");

	InitTriangulation(Nh,Nv,&back,&fore);

	for (y=0; y<Nv; y++) {
	    for (x=0; x<Nh; x++) {
		back[y][x]=0; fore[y][x]=(long)cell[y][x];
	    }
	    free(cell[y]);
	}
	free(cell);

#define ALLOC(type,noe)	(type *)malloc(sizeof(type)*(noe))

	if ((C=ALLOC(double,(size_t)M))==NULL ||
	    (S=ALLOC(double,(size_t)M))==NULL ||
	    (P=ALLOC(double *,(size_t)M))==NULL ||
	    (*P=ALLOC(double,(size_t)M*(size_t)N))==NULL)
	    Error("memory allocation error.");

	for (y=0; y<M; y++) {
	    t=M_PI*(double)y/(double)M; C[y]=cos(t); S[y]=sin(t);

	    if (y>0) P[y]=P[y-1]+N;

	    for (x=0; x<N; x++) P[y][x]=0.0;
	}
	(void)fprintf(stderr,"%d\t%d\t%s\t%lg\t",N,M,argv[4],(double)(1-N)/2.0);

	Triangulation(0,0,Nh-1,Nv-1,back,fore,Store);

	for (y=0; y<M; y++)
	for (x=0; x<N; x++)
	    if (x==0 && y==0)
		p1=p2=P[y][x];
	    else
		if (P[y][x]<p1) p1=P[y][x]; else if (P[y][x]>p2) p2=P[y][x];

	(void)fprintf(stderr,"%lf\t%lf\n",p1,p2);

	if (argv[argc-1][0]=='-')
	    file=stdout;
	else
	    if ((file=fopen(argv[argc-1],"wb"))==NULL) Error("file not open.");

	(void)fprintf(file,"%d\t%d\n",N,M);
	(void)fwrite(*P,sizeof(double),(size_t)M*(size_t)N,file);

	if (file!=stdout) (void)fclose(file);
/*
	free(*P); free(P); free(S); free(C); TermTriangulation(back,fore);
*/
	return 0;
}
