
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "rif.h"
#include "t_m.h"

#ifdef  WINDOWS
#include <windows.h>

#define THREAD_VAR	HANDLE
#define THREAD_FUNC	DWORD WINAPI
#define THREAD_ARG	LPVOID
#define THREAD_NULL	0
#define PIPE_VAR	HANDLE

#define THREAD_CREATE(var,func,arg)	\
	(var=CreateThread(NULL,0,func,(THREAD_ARG)(arg),0,NULL))==0
#define THREAD_JOIN(var)	\
	WaitForSingleObject(var,INFINITE)==WAIT_FAILED || CloseHandle(var)==0

#define PIPE_OPEN(ptr)	(CreatePipe(ptr,ptr+1,NULL,0)==0)
#define PIPE_CLOSE(var)	CloseHandle(var)==0

static int	Read(var,ptr,cnt)
HANDLE		var;
LPVOID		ptr;
DWORD		cnt;
{
	DWORD	num;

	return (ReadFile(var,ptr,cnt,&num,NULL))?num:
	       (GetLastError()==ERROR_BROKEN_PIPE)?0:-1;
}

static int	Write(var,ptr,cnt)
HANDLE		var;
LPVOID		ptr;
DWORD		cnt;
{
	DWORD	num;

	return (WriteFile(var,ptr,cnt,&num,NULL))?num:
	       (GetLastError()==ERROR_BROKEN_PIPE)?0:-1;
}

#define PIPE_READ(var,ptr,cnt)	Read(var,(LPVOID)ptr,(DWORD)cnt)
#define PIPE_WRITE(var,ptr,cnt)	Write(var,(LPVOID)ptr,(DWORD)cnt)
#else
#include <pthread.h>
#include <unistd.h>

#define THREAD_VAR	pthread_t
#define THREAD_FUNC	void *
#define THREAD_ARG	void *
#define THREAD_NULL	NULL
#define PIPE_VAR	int

#define THREAD_CREATE(var,func,arg)	\
	pthread_create(&(var),NULL,func,(THREAD_ARG)(arg))
#define THREAD_JOIN(var)	pthread_join(var,NULL)

#define PIPE_OPEN(ptr)	pipe(ptr)
#define PIPE_CLOSE(var)	close(var)

#define PIPE_READ(var,ptr,cnt)	(int)read(var,(void *)ptr,(size_t)cnt)
#define PIPE_WRITE(var,ptr,cnt)	(int)write(var,(void *)ptr,(size_t)cnt)
#endif

#ifndef	THREADS
#define THREADS	1
#endif

extern void	Error();

#ifndef	M_PI
#define M_PI	3.14159265358979323846
#endif

static PIPE_VAR	*I;
static int	N,M,L;
static double	R,*C,*S,**P;

typedef	struct {
		double	x1,y1,x2,y2,x3,y3,f;
	} Data;

#define SOD	sizeof(Data)

static THREAD_FUNC	Thread(l)
THREAD_ARG		l;
{
	Data	d;
	int	i,n2,n;
	size_t	m;
	double	r1,s1,r2,s2,r3,s3,r,s,t31,t,*p;

	while ((i=PIPE_READ(I[(size_t)l],&d,SOD))==SOD)
	    for (m=(size_t)l; m<M; m+=L) {
		r1=R+C[m]*d.x1+S[m]*d.y1; s1=C[m]*d.y1-S[m]*d.x1;
		r2=R+C[m]*d.x2+S[m]*d.y2; s2=C[m]*d.y2-S[m]*d.x2;
		r3=R+C[m]*d.x3+S[m]*d.y3; s3=C[m]*d.y3-S[m]*d.x3;
		if (r1>r2)
		    if (r3>r1) {
			r=r1; r1=r2; r2=r;
			s=s1; s1=s2; s2=s;
		    }
		    else if (r3>r2) {
			r=r1; r1=r2; r2=r3; r3=r;
			s=s1; s1=s2; s2=s3; s3=s;
		    }
		    else {
			r=r3; r3=r1; r1=r;
			s=s3; s3=s1; s1=s;
		    }
		else
		    if (r1>r3) {
			r=r3; r3=r2; r2=r1; r1=r;
			s=s3; s3=s2; s2=s1; s1=s;
		    }
		    else if (r2>r3) {
			r=r2; r2=r3; r3=r;
			s=s2; s2=s3; s3=s;
		    }

		if (r1>=(double)N || r3<0.0) continue;

		n2=(int)floor(r2); t31=(s3-s1)/(r3-r1); p=P[m];

		if (n2>=0) {
		    r=r2-r1;
		    if ((n=(int)floor(r1))==n2)
			p[n2]+=(d.f*fabs(s2-s1-t31*r)*r);
		    else {
			t=d.f*fabs((s2-s1)/r-t31);
			if (n<0) {
			    s1=t*(-r1); n=0;
			}
			else {
			    r1=(double)(n+1)-r1; s1=t*r1; p[n++]+=(s1*r1);
			}
			while (n<n2 && n<N) {
			    s=s1; s1+=t; p[n++]+=(s+s1);
			}
			if (n<N) p[n2]+=((s1+t*r)*(r2-(double)n2));
		    }
		}
		if (n2<N) {
		    r=r3-r2;
		    if ((n=(int)floor(r3))==n2)
			p[n2]+=(d.f*fabs(t31*r-s3+s2)*r);
		    else {
			t=d.f*fabs(t31-(s3-s2)/r);
			if (n>=N) {
			    s3=t*(r3-(double)N); n=N-1;
			}
			else {
			    r3-=(double)n; s3=t*r3; p[n--]+=(s3*r3);
			}
			while (n>n2 && n>=0) {
			    s=s3; s3+=t; p[n--]+=(s+s3);
			}
			if (n>=0) p[n2]+=((s3+t*r)*((double)(n2+1)-r2));
		    }
		}
	    }

	if (i) Error("pipe read error.");

	return THREAD_NULL;
}

static PIPE_VAR	*O;
static double	f0,df,h0,v0;

static void	Store(h1,v1,h2,v2,h3,v3,c)
int		h1,v1,h2,v2,h3,v3;
long		c;
{
	Data	d;
	int	l;

	d.x1=(double)h1-h0; d.y1=v0-(double)v1;
	d.x2=(double)h2-h0; d.y2=v0-(double)v2;
	d.x3=(double)h3-h0; d.y3=v0-(double)v3; d.f=f0+df*(double)c;
	for (l=0; l<L; l++)
	    if (PIPE_WRITE(O[l],&d,SOD)!=SOD) Error("pipe write error.");
}

int	main(argc,argv)
int	argc;
char	**argv;
{
	int		Nh,Nv,BPS,y,x,l;
	Cell		**cell;
	double		dr,t,p1,p2;
	char		*env,**back;
	long		**fore;
	PIPE_VAR	io[2];
	THREAD_VAR	*T;
	FILE		*file;

	if (argc!=6 && argc!=7)
	    Error("usage : tg2sg tomogram f1 f2 dr {M} sinogram");

	ReadImageFile(argv[1],&Nh,&Nv,&BPS,&cell,NULL);

	if (Nh!=Nv)
	    (void)fputs("bad image size as a tomogram (warning).\n",stderr);

	dr=0.5*atof(argv[4]);
	f0=dr*atof(argv[2]);
	df=(dr*atof(argv[3])-f0)/(double)((1<<BPS)-1);

	N=(Nh<Nv)?Nh:Nv; R=(double)N/2.0; h0=(double)Nh/2.0; v0=(double)Nv/2.0;

	if (argc==6)
	    M=(int)ceil(M_PI*R);
	else
	    if ((M=atoi(argv[5]))<=0) Error("bad number of projections.");

	L=((env=getenv("THREADS"))==NULL)?THREADS:atoi(env);
	if (L<=0 || L>M) Error("bad number of THREADS.");

	InitTriangulation(Nh,Nv,&back,&fore);

	for (y=0; y<Nv; y++) {
	    for (x=0; x<Nh; x++) {
		back[y][x]=0; fore[y][x]=(long)cell[y][x];
	    }
	    free(cell[y]);
	}
	free(cell);

#define ALLOC(type,noe)	(type *)malloc(sizeof(type)*(noe))

	if ((C=ALLOC(double,(size_t)M))==NULL ||
	    (S=ALLOC(double,(size_t)M))==NULL ||
	    (P=ALLOC(double *,(size_t)M))==NULL ||
	    (*P=ALLOC(double,(size_t)M*(size_t)N))==NULL ||
	    (I=ALLOC(PIPE_VAR,L))==NULL ||
	    (O=ALLOC(PIPE_VAR,L))==NULL ||
	    (T=ALLOC(THREAD_VAR,L))==NULL) Error("memory allocation error.");

	for (y=0; y<M; y++) {
	    t=M_PI*(double)y/(double)M; C[y]=cos(t); S[y]=sin(t);

	    if (y>0) P[y]=P[y-1]+N;

	    for (x=0; x<N; x++) P[y][x]=0.0;
	}
	(void)fprintf(stderr,"%d\t%d\t%s\t%lg\t",N,M,argv[4],(double)(1-N)/2.0);

#ifdef	PTW32_STATIC_LIB
	if (!pthread_win32_process_attach_np())
	    Error("pthreads-win32 initialization failed.");
#endif
	for (l=0; l<L; l++) {
	    if (PIPE_OPEN(io)) Error("can not open pipe.");

	    I[l]=io[0]; O[l]=io[1];

	    if (THREAD_CREATE(T[l],Thread,l)) Error("can not create thread.");
	}
	Triangulation(0,0,Nh-1,Nv-1,back,fore,Store);

	for (l=0; l<L; l++) {
	    if (PIPE_CLOSE(O[l]) || THREAD_JOIN(T[l]) || PIPE_CLOSE(I[l]))
		Error("can not close pipe or can not join thread.");
	}
#ifdef	PTW32_STATIC_LIB
	(void)pthread_win32_process_detach_np();
#endif
	for (y=0; y<M; y++)
	for (x=0; x<N; x++)
	    if (x==0 && y==0)
		p1=p2=P[y][x];
	    else
		if (P[y][x]<p1) p1=P[y][x]; else if (P[y][x]>p2) p2=P[y][x];

	(void)fprintf(stderr,"%lf\t%lf\n",p1,p2);

	if (argv[argc-1][0]=='-')
	    file=stdout;
	else
	    if ((file=fopen(argv[argc-1],"wb"))==NULL) Error("file not open.");

	(void)fprintf(file,"%d\t%d\n",N,M);
	(void)fwrite(*P,sizeof(double),(size_t)M*(size_t)N,file);

	if (file!=stdout) (void)fclose(file);
/*
	free(T); free(O); free(I); free(*P); free(P); free(S); free(C);
	TermTriangulation(back,fore);
*/
	return 0;
}
