
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif.h"
#include "csi.h"
#include "sif.h"

#ifdef	__MSVCRT__
#define LL_FORMAT	"I64"
#else
#define LL_FORMAT	"ll"
#endif

extern void	Error();

#define STR_LEN	1024
#define BPS	(8*sizeof(Cell))
#define SPRINTF	(void)sprintf
#define FPRINTF	(void)fprintf

int	main(argc,argv)
int	argc;
char	**argv;
{
	char			**name,path[STR_LEN],*Desc,
				val[8][STR_LEN],BS[STR_LEN],*ptr,desc[STR_LEN];
	int			Nz,z,BpS,bs,PV,pv,Nx,Ny,y,x,bps=0;
	double			F0,dF,F,f0,df,f;
	unsigned long long	pels[1<<BPS];
	Cell			**cell;

	if (argc!=3 && argc!=4 && argc!=6 && argc!=7)
	    Error("usage : tg2tg srcDir nameFile {f0 df BPS} {dstDir}");

	name=CheckSliceImages(argv[1],argv[2],NULL,NULL,&Nz,NULL);

	if (argc==3 || argc==4) {
	    for (z=0; z<Nz; z++) {
		SPRINTF(path,"%s/%s",argv[1],name[z]);
		ReadImageFile(path,NULL,NULL,&BpS,NULL,&Desc);

		if (Desc!=NULL) {
		    switch(sscanf(Desc,"%s %s %s %s %s %s %s %s",
				       val[0],val[1],val[2],
				       val[3],val[4],val[5],val[6],val[7])) {
			case 2 :
			    bs=sscanf(val[0],"%lf",&F0)==1 &&
			       sscanf(val[1],"%lf",&F )==1; break;
			case 3 :
			    if (bs=sscanf(val[1],"%lf",&F0)==1 &&
				   sscanf(val[2],"%lf",&dF)==1)
				F=F0+dF*(double)((1<<BpS)-1);

			    break;
			case 4 :
			    if (bs=sscanf(val[2],"%lf",&F0)==1 &&
				   sscanf(val[3],"%lf",&dF)==1)
				F=F0+dF*(double)((1<<BpS)-1);

			    break;
			case 6 :
			    bs=sscanf(val[4],"%lf",&F0)==1 &&
			       sscanf(val[5],"%lf",&F )==1; break;
			case 8 :
			    if (bs=sscanf(val[6],"%lf",&F0)==1 &&
				   sscanf(val[7],"%lf",&dF)==1)
				F=F0+dF*(double)((1<<BpS)-1);

			    break;
			default:
			    bs=0; break;
		    }
		    free(Desc);

		    if (bs!=0)
			if (bps==0) {
			    bps=BpS; f0=F0; f=F;
			}
			else {
			    if (BpS>bps) bps=BpS;
			    if (F0<f0) f0=F0;
			    if (F>f) f=F;
			}
		}
	    }
	    if (bps==0 || f0>=f) Error("no or bad image description(s).");

	    FPRINTF(stderr,"%lf\t%lf\n",f0,f);

	    df=(f-f0)/(double)(PV=(1<<bps)-1);

	    if (argc==4) {
		SPRINTF(BS,"%lf\t%20.18lf",f0,df);
		for (ptr=BS; *ptr!='\0'; ptr++) ;
		while (*(--ptr)=='0') *ptr='\0';
		if (*ptr=='.') *ptr='\0';
	    }
	}
	else {
	    f0=atof(argv[3]);
	    if ((df=atof(argv[4]))<=0.0) Error("bad CT-value step.");

	    if ((bps=atoi(argv[5]))<=0 || bps>BPS) Error("bad BPS.");

	    f=f0+df*(double)(PV=(1<<bps)-1);

	    if (argc==7) SPRINTF(BS,"%s\t%s",argv[3],argv[4]);
	}
	f0-=(df/2.0);

	for (pv=0; pv<=PV; pv++) pels[pv]=0;

	for (z=0; z<Nz; z++) {
	    SPRINTF(path,"%s/%s",argv[1],name[z]);
	    ReadImageFile(path,&Nx,&Ny,&BpS,&cell,&Desc);

	    if (Desc==NULL)
		FPRINTF(stderr,"%s : no image description.\n",name[z]);
	    else {
		switch(sscanf(Desc,"%s %s %s %s %s %s %s %s",
				   val[0],val[1],val[2],
				   val[3],val[4],val[5],val[6],val[7])) {
		    case 2 :
			if (bs=(sscanf(val[0],"%lf",&F0)==1 &&
				sscanf(val[1],"%lf",&F )==1)?2:0)
			    dF=(F-F0)/(double)((1<<BpS)-1);

			break;
		    case 3 :
			bs=(sscanf(val[1],"%lf",&F0)==1 &&
			    sscanf(val[2],"%lf",&dF)==1)?3:0; break;
		    case 4 :
			bs=(sscanf(val[2],"%lf",&F0)==1 &&
			    sscanf(val[3],"%lf",&dF)==1)?4:0; break;
		    case 6 :
			if (bs=(sscanf(val[4],"%lf",&F0)==1 &&
				sscanf(val[5],"%lf",&F )==1)?6:0)
			    dF=(F-F0)/(double)((1<<BpS)-1);

			break;
		    case 8 :
			bs=(sscanf(val[6],"%lf",&F0)==1 &&
			    sscanf(val[7],"%lf",&dF)==1)?8:0; break;
		    default:
			bs=0; break;
		}
		free(Desc);

		if (bs==0)
		    FPRINTF(stderr,"%s : bad image description.\n",name[z]);
		else {
		    for (y=0; y<Ny; y++)
		    for (x=0; x<Nx; x++)
			++pels[cell[y][x]=((F=F0+dF*(double)cell[y][x])<f0)?0:
					  (F<f)?(Cell)((F-f0)/df):
					  (Cell)PV];

		    if (argc==4 || argc==7) {
			switch(bs) {
			    case 2 :
			    case 4 :
				SPRINTF(desc,"%s\t%s\n%s",val[0],val[1],BS);
				break;
			    case 3 :
				SPRINTF(desc,"%s\t%s",val[0],BS); break;
			    case 6 :
			    case 8 :
				SPRINTF(desc,"%s\t%s\n%s\t%s\n%s\t%s\n%s",
					     val[0],val[1],val[2],
					     val[3],val[4],val[5],BS);
				break;
			}
			SPRINTF(path,"%s/%s",argv[argc-1],name[z]);
			StoreImageFile(path,Nx,Ny,bps,cell,desc);
		    }
		}
	    }
	    for (y=Ny-1; y>=0; y--) free(cell[y]); free(cell);
	}
	for (pv=0; pv<=PV; pv++)
	    if (pels[pv])
		(void)printf("%d\t%lf\t%lf\t%" LL_FORMAT "u\n",
			     pv,f0+df*(double)pv,f0+df*(double)(pv+1),pels[pv]);

	return 0;
}
